import logging

from pathlib import Path
from typing import Dict, Sequence, Union

import bakta
import bakta.config as cfg
import bakta.constants as bc
import bakta.io.fasta as fasta
import bakta.io.insdc as insdc
import bakta.features.annotation as ba
import bakta.so as so


log = logging.getLogger('GFF')


def write_features(data: dict, features_by_sequence: Dict[str, dict], gff3_path: Path):
    """Export features in GFF3 format."""
    log.info('write features: path=%s', gff3_path)

    with gff3_path.open('wt') as fh:
        fh.write('##gff-version 3\n')  # GFF version
        fh.write('##feature-ontology https://github.com/The-Sequence-Ontology/SO-Ontologies/blob/v3.1/so.obo\n')  # SO feature version

        if(data['genome'].get('taxon', None)):  # write organism info
            fh.write(f"# organism {data['genome']['taxon']}\n")

        fh.write('# Annotated with Bakta\n')
        fh.write(f'# Software: v{cfg.version}\n')
        fh.write(f"# Database: v{cfg.db_info['major']}.{cfg.db_info['minor']}, {cfg.db_info['type']}\n")
        fh.write(f'# DOI: {bc.BAKTA_DOI}\n')
        fh.write(f'# URL: {bc.BAKTA_URL}\n')

        for seq in data['sequences']:  # write features
            fh.write(f"##sequence-region {seq['id']} 1 {seq['length']}\n")  # sequence region

            # write landmark region
            annotations = {
                'ID': seq['id'],
                'Name': seq['id']
            }
            if(seq['topology'] == bc.TOPOLOGY_CIRCULAR):
                annotations['Is_circular'] = 'true'
            annotations = encode_annotations(annotations)
            fh.write(f"{seq['id']}\tBakta\tregion\t1\t{str(seq['length'])}\t.\t+\t.\t{annotations}\n")

            for feat in features_by_sequence[seq['id']]:
                seq_id = feat['sequence'] if 'sequence' in feat else feat['contig']  # <1.10.0 compatibility
                start = feat['start']
                stop = feat['stop']
                if('edge' in feat):
                    stop += seq['length']

                if(feat['type'] == bc.FEATURE_T_RNA):
                    annotations = {
                        'ID': feat['locus'],
                        'Name': feat['product'],
                        'locus_tag': feat['locus'],
                        'product': feat['product'],
                        'Dbxref': feat['db_xrefs']
                    }
                    if(feat.get('gene', None)):  # add gene annotation if available
                        annotations['gene'] = feat['gene']
                    if(bc.PSEUDOGENE in feat):
                        annotations[bc.INSDC_FEATURE_PSEUDOGENE] = bc.INSDC_FEATURE_PSEUDOGENE_TYPE_UNKNOWN
                    elif('truncated' in feat):
                        annotations[bc.INSDC_FEATURE_PSEUDO] = True
                    if(feat.get('anti_codon', False)):
                        annotations['anti_codon'] = feat['anti_codon']
                    if(feat.get('amino_acid', False)):
                        annotations['amino_acid'] = feat['amino_acid']
                    if(cfg.compliant):
                        gene_id = f"{feat['locus']}_gene"
                        annotations['Parent'] = gene_id
                        annotations['inference'] = 'profile:tRNAscan:2.0'
                        annotations['Dbxref'], annotations['Note'] = insdc.revise_dbxref_insdc(feat['db_xrefs'])  # remove INSDC invalid DbXrefs
                        gene_annotations = {
                            'ID': gene_id,
                            'locus_tag': feat['locus']
                        }
                        if(feat.get('gene', None)):
                            gene_annotations['gene'] = feat['gene']
                        if(bc.PSEUDOGENE in feat):
                            gene_annotations[bc.INSDC_FEATURE_PSEUDOGENE] = bc.INSDC_FEATURE_PSEUDOGENE_TYPE_UNKNOWN
                        gene_annotations = encode_annotations(gene_annotations)
                        fh.write(f"{seq_id}\ttRNAscan-SE\tgene\t{start}\t{stop}\t.\t{feat['strand']}\t.\t{gene_annotations}\n")
                    annotations = encode_annotations(annotations)
                    fh.write(f"{seq_id}\ttRNAscan-SE\t{so.SO_TRNA.name}\t{start}\t{stop}\t.\t{feat['strand']}\t.\t{annotations}\n")
                elif(feat['type'] == bc.FEATURE_TM_RNA):
                    annotations = {
                        'ID': feat['locus'],
                        'Name': feat['product'],
                        'locus_tag': feat['locus'],
                        'gene': feat['gene'],
                        'product': feat['product'],
                        'Dbxref': feat['db_xrefs']
                    }
                    if('tag' in feat):
                        annotations['tag_peptide'] = feat['tag']['aa']
                    if('truncated' in feat):
                        annotations[bc.INSDC_FEATURE_PSEUDO] = True
                    if(cfg.compliant):
                        gene_id = f"{feat['locus']}_gene"
                        annotations['Parent'] = gene_id
                        annotations['inference'] = 'profile:aragorn:1.2'
                        annotations['Dbxref'], annotations['Note'] = insdc.revise_dbxref_insdc(feat['db_xrefs'])  # remove INSDC invalid DbXrefs
                        if('tag' in feat):
                            annotations['tag_peptide'] = f"{feat['tag']['start']}..{feat['tag']['stop']}" if feat['strand'] == bc.STRAND_FORWARD else f"complement({feat['tag']['start']}..{feat['tag']['stop']})"
                        gene_annotations = {
                            'ID': gene_id,
                            'locus_tag': feat['locus'],
                            'gene': feat['gene']
                        }
                        if('truncated' in feat):
                            gene_annotations[bc.INSDC_FEATURE_PSEUDO] = True
                        gene_annotations = encode_annotations(gene_annotations)
                        fh.write(f"{seq_id}\tAragorn\tgene\t{start}\t{stop}\t.\t{feat['strand']}\t.\t{gene_annotations}\n")
                    annotations = encode_annotations(annotations)
                    fh.write(f"{seq_id}\tAragorn\t{so.SO_TMRNA.name}\t{start}\t{stop}\t.\t{feat['strand']}\t.\t{annotations}\n")
                elif(feat['type'] == bc.FEATURE_R_RNA):
                    annotations = {
                        'ID': feat['locus'],
                        'Name': feat['product'],
                        'locus_tag': feat['locus'],
                        'gene': feat['gene'],
                        'product': feat['product'],
                        'Dbxref': feat['db_xrefs']
                    }
                    if('truncated' in feat):
                        annotations[bc.INSDC_FEATURE_PSEUDO] = True
                    if(cfg.compliant):
                        gene_id = f"{feat['locus']}_gene"
                        annotations['Parent'] = gene_id
                        annotations['Dbxref'], annotations['Note'] = insdc.revise_dbxref_insdc(feat['db_xrefs'])  # remove INSDC invalid DbXrefs
                        for rfam_id in [dbxref.split(':')[1] for dbxref in feat['db_xrefs'] if dbxref.split(':')[0] == bc.DB_XREF_RFAM]:
                            annotations['inference'] = f'profile:Rfam:{rfam_id}'
                        gene_annotations = {
                            'ID': gene_id,
                            'locus_tag': feat['locus'],
                            'gene': feat['gene']
                        }
                        if('truncated' in feat):
                            gene_annotations[bc.INSDC_FEATURE_PSEUDO] = True
                        gene_annotations = encode_annotations(gene_annotations)
                        fh.write(f"{seq_id}\tInfernal\tgene\t{start}\t{stop}\t.\t{feat['strand']}\t.\t{gene_annotations}\n")
                    annotations = encode_annotations(annotations)
                    fh.write(f"{seq_id}\tInfernal\t{so.SO_RRNA.name}\t{start}\t{stop}\t{feat['evalue']}\t{feat['strand']}\t.\t{annotations}\n")
                elif(feat['type'] == bc.FEATURE_NC_RNA):
                    annotations = {
                        'ID': feat['locus'],
                        'Name': feat['product'],
                        'locus_tag': feat['locus'],
                        'gene': feat['gene'],
                        'product': feat['product'],
                        'Dbxref': feat['db_xrefs']
                    }
                    if('truncated' in feat):
                        annotations[bc.INSDC_FEATURE_PSEUDO] = True
                    if(cfg.compliant):
                        gene_id = f"{feat['locus']}_gene"
                        annotations['Parent'] = gene_id
                        annotations['Dbxref'], annotations['Note'] = insdc.revise_dbxref_insdc(feat['db_xrefs'])  # remove INSDC invalid DbXrefs
                        annotations[bc.INSDC_FEATURE_NC_RNA_CLASS] = insdc.select_ncrna_class(feat)
                        for rfam_id in [dbxref.split(':')[1] for dbxref in feat['db_xrefs'] if dbxref.split(':')[0] == bc.DB_XREF_RFAM]:
                            annotations['inference'] = f'profile:Rfam:{rfam_id}'
                        gene_annotations = {
                            'ID': gene_id,
                            'locus_tag': feat['locus'],
                            'gene': feat['gene']
                        }
                        if(ba.RE_GENE_SYMBOL.fullmatch(feat['gene'])):  # discard non-standard ncRNA gene symbols
                            gene_annotations['gene'] = feat['gene']
                        else:
                            annotations.pop('gene', None)
                        if('truncated' in feat):
                            gene_annotations[bc.INSDC_FEATURE_PSEUDO] = True
                        gene_annotations = encode_annotations(gene_annotations)
                        fh.write(f"{seq_id}\tInfernal\tgene\t{start}\t{stop}\t.\t{feat['strand']}\t.\t{gene_annotations}\n")
                    annotations = encode_annotations(annotations)
                    fh.write(f"{seq_id}\tInfernal\t{so.SO_NCRNA_GENE.name}\t{start}\t{stop}\t{feat['evalue']}\t{feat['strand']}\t.\t{annotations}\n")
                elif(feat['type'] == bc.FEATURE_NC_RNA_REGION):
                    annotations = {
                        'ID': feat['id'],
                        'Name': feat['product'],
                        'product': feat['product'],
                        'Dbxref': feat['db_xrefs']
                    }
                    if('truncated' in feat):
                        annotations[bc.INSDC_FEATURE_PSEUDO] = True
                    if(cfg.compliant):
                        for rfam_id in [dbxref.split(':')[1] for dbxref in feat['db_xrefs'] if dbxref.split(':')[0] == bc.DB_XREF_RFAM]:
                            annotations['inference'] = f'profile:Rfam:{rfam_id}'
                        annotations['Dbxref'], annotations['Note'] = insdc.revise_dbxref_insdc(feat['db_xrefs'])  # remove INSDC invalid DbXrefs
                        annotations[bc.INSDC_FEATURE_REGULATORY_CLASS] = insdc.select_regulatory_class(feat)
                    annotations = encode_annotations(annotations)
                    fh.write(f"{seq_id}\tInfernal\t{so.SO_REGULATORY_REGION.name}\t{start}\t{stop}\t{feat['evalue']}\t{feat['strand']}\t.\t{annotations}\n")
                elif(feat['type'] == bc.FEATURE_CRISPR):
                    annotations = {
                        'ID': feat['id'],
                        'Name': feat['product'],
                        'product': feat['product']
                    }
                    feat_type = so.SO_CRISPR.name
                    if(cfg.compliant):
                        feat_type = bc.INSDC_FEATURE_REPEAT_REGION
                        annotations['inference'] = 'COORDINATES:alignment:pilercr:1.02'
                        annotations['Dbxref'], annotations['Note'] = insdc.revise_dbxref_insdc(feat['db_xrefs'])  # remove INSDC invalid DbXrefs
                        annotations[bc.INSDC_FEATURE_REPEAT_FAMILY] = 'CRISPR'
                        annotations[bc.INSDC_FEATURE_REPEAT_TYPE] = 'direct'
                        annotations[bc.INSDC_FEATURE_REPEAT_UNIT_SEQ] = feat['repeat_consensus']
                    annotations = encode_annotations(annotations)
                    fh.write(f"{seq_id}\tPILER-CR\t{feat_type}\t{start}\t{stop}\t.\t{feat['strand']}\t.\t{annotations}\n")
                    if(not cfg.compliant):
                        i = 0
                        while i < len(feat['spacers']):
                            repeat = feat['repeats'][i]
                            annotations = {
                                'ID': f"{feat['id']}_repeat_{i+1}",
                                'Parent': feat['id']
                            }
                            annotations = encode_annotations(annotations)
                            fh.write(f"{seq_id}\tPILER-CR\t{bc.FEATURE_CRISPR_REPEAT}\t{repeat['start']}\t{repeat['stop']}\t.\t{repeat['strand']}\t.\t{annotations}\n")
                            spacer = feat['spacers'][i]
                            annotations = {
                                'ID': f"{feat['id']}_spacer_{i+1}",
                                'Parent': feat['id'],
                                'sequence': spacer['sequence']
                            }
                            annotations = encode_annotations(annotations)
                            fh.write(f"{seq_id}\tPILER-CR\t{bc.FEATURE_CRISPR_SPACER}\t{spacer['start']}\t{spacer['stop']}\t.\t{spacer['strand']}\t.\t{annotations}\n")
                            i += 1
                        if(len(feat['repeats']) - 1 == i):
                            repeat = feat['repeats'][i]
                            annotations = { 'ID': f"{feat['id']}_repeat_{i+1}" }
                            annotations = encode_annotations(annotations)
                            fh.write(f"{seq_id}\tPILER-CR\t{bc.FEATURE_CRISPR_REPEAT}\t{repeat['start']}\t{repeat['stop']}\t.\t{repeat['strand']}\t.\t{annotations}\n")
                elif(feat['type'] == bc.FEATURE_CDS):
                    annotations = {
                        'ID': feat['locus'],
                        'Name': feat['product'],
                        'locus_tag': feat['locus'],
                        'product': feat['product'],
                        'Dbxref': feat['db_xrefs']
                    }
                    if(bc.PSEUDOGENE in feat):
                        annotations[bc.INSDC_FEATURE_PSEUDOGENE] = bc.INSDC_FEATURE_PSEUDOGENE_TYPE_UNPROCESSED if feat[bc.PSEUDOGENE]['paralog'] else bc.INSDC_FEATURE_PSEUDOGENE_TYPE_UNITARY
                    elif('truncated' in feat):
                        annotations[bc.INSDC_FEATURE_PSEUDO] = True
                    if(feat.get('gene', None)):  # add gene annotation if available
                        annotations['gene'] = feat['gene']
                    source = '?' if feat.get('source', None) == bc.CDS_SOURCE_USER else 'Pyrodigal'
                    if(cfg.compliant):
                        gene_id = f"{feat['locus']}_gene"
                        annotations['Parent'] = gene_id
                        annotations['inference'] = 'EXISTENCE:non-experimental evidence, no additional details recorded' if feat.get('source', None) == bc.CDS_SOURCE_USER else 'ab initio prediction:Pyrodigal:3.5'
                        annotations['Dbxref'], annotations['Note'] = insdc.revise_dbxref_insdc(feat['db_xrefs'])  # remove INSDC invalid DbXrefs
                        annotations['Note'], ec_number = insdc.extract_ec_from_notes_insdc(annotations, 'Note')
                        if(ec_number is not None):
                            annotations['ec_number'] = ec_number
                        gene_annotations = {
                            'ID': gene_id,
                            'locus_tag': feat['locus']
                        }
                        if(feat.get('gene', None)):
                            gene_annotations['gene'] = feat['gene']
                        if(bc.PSEUDOGENE in feat):
                            gene_annotations[bc.INSDC_FEATURE_PSEUDOGENE] = bc.INSDC_FEATURE_PSEUDOGENE_TYPE_UNPROCESSED if feat[bc.PSEUDOGENE]['paralog'] else bc.INSDC_FEATURE_PSEUDOGENE_TYPE_UNITARY
                        gene_annotations = encode_annotations(gene_annotations)
                        fh.write(f"{seq_id}\t{source}\tgene\t{start}\t{stop}\t.\t{feat['strand']}\t.\t{gene_annotations}\n")
                    if('exception' in feat):
                        ex = feat['exception']
                        if(ex['type'] == 'selenocysteine'):
                            pos = f"{ex['start']}..{ex['stop']}"
                            if(feat['strand'] == bc.STRAND_REVERSE):
                                pos = f"complement({pos})"
                            annotations['transl_except']=f"(pos:{pos},aa:{ex['aa']})"
                            notes = annotations.get('Note', [])
                            notes.append(f"codon on position {ex['codon_position']} is a {ex['type']} codon")
                            if('Notes' not in annotations):
                                annotations['Note'] = notes
                        elif(ex['type'] == 'ribosomal_slippage'):
                            annotations[bc.INSDC_FEAUTRE_CDS_RIBOSOMAL_SLIPPAGE] = True
                    annotations = encode_annotations(annotations)
                    fh.write(f"{seq_id}\t{source}\t{so.SO_CDS.name}\t{start}\t{stop}\t.\t{feat['strand']}\t0\t{annotations}\n")
                    if(bc.FEATURE_SIGNAL_PEPTIDE in feat):
                        write_signal_peptide(fh, feat)
                elif(feat['type'] == bc.FEATURE_SORF):
                    annotations = {
                        'ID': feat['locus'],
                        'Name': feat['product'],
                        'locus_tag': feat['locus'],
                        'product': feat['product'],
                        'Dbxref': feat['db_xrefs']
                    }
                    if(feat.get('gene', None)):  # add gene annotation if available
                        annotations['gene'] = feat['gene']
                    if(cfg.compliant):
                        gene_id = f"{feat['locus']}_gene"
                        annotations['Parent'] = gene_id
                        annotations['Dbxref'], annotations['Note'] = insdc.revise_dbxref_insdc(feat['db_xrefs'])  # remove INSDC invalid DbXrefs
                        annotations['Note'], ec_number = insdc.extract_ec_from_notes_insdc(annotations, 'Note')
                        if(ec_number is not None):
                            annotations['ec_number'] = ec_number
                        gene_annotations = {
                            'ID': gene_id,
                            'locus_tag': feat['locus'],
                            'inference': 'ab initio prediction:Bakta'
                        }
                        if(feat.get('gene', None)):
                            gene_annotations['gene'] = feat['gene']
                        gene_annotations = encode_annotations(gene_annotations)
                        fh.write(f"{seq_id}\tBakta\tgene\t{start}\t{stop}\t.\t{feat['strand']}\t.\t{gene_annotations}\n")
                    annotations = encode_annotations(annotations)
                    fh.write(f"{seq_id}\tBakta\t{so.SO_CDS.name}\t{start}\t{stop}\t.\t{feat['strand']}\t0\t{annotations}\n")
                    if(bc.FEATURE_SIGNAL_PEPTIDE in feat):
                        write_signal_peptide(fh, feat)
                elif(feat['type'] == bc.FEATURE_GAP):
                    annotations = {
                        'ID': feat['id'],
                        'Name': f"gap ({feat['length']} bp)",
                        'product': f"gap ({feat['length']} bp)"
                    }
                    annotations = encode_annotations(annotations)
                    fh.write(f"{seq_id}\tBakta\t{so.SO_GAP.name}\t{start}\t{stop}\t.\t{feat['strand']}\t.\t{annotations}\n")
                elif(feat['type'] == bc.FEATURE_ORIC):
                    annotations = {
                        'ID': feat['id'],
                        'Name': feat['product']
                    }
                    if(cfg.compliant):
                        annotations['Note'] = feat['product']
                    else:
                        annotations['product'] = feat['product']
                        annotations['inference'] = 'similar to DNA sequence'
                    annotations = encode_annotations(annotations)
                    feat_type = bc.INSDC_FEATURE_ORIGIN_REPLICATION if cfg.compliant else so.SO_ORIC.name
                    fh.write(f"{seq_id}\tBLAST+\t{feat_type}\t{start}\t{stop}\t.\t{feat['strand']}\t.\t{annotations}\n")
                elif(feat['type'] == bc.FEATURE_ORIV):
                    annotations = {
                        'ID': feat['id'],
                        'Name': feat['product']
                    }
                    if(cfg.compliant):
                        annotations['Note'] = feat['product']
                    else:
                        annotations['product'] = feat['product']
                        annotations['inference'] = 'similar to DNA sequence'
                    annotations = encode_annotations(annotations)
                    feat_type = bc.INSDC_FEATURE_ORIGIN_REPLICATION if cfg.compliant else so.SO_ORIC.name
                    fh.write(f"{seq_id}\tBLAST+\t{feat_type}\t{start}\t{stop}\t.\t{feat['strand']}\t.\t{annotations}\n")
                elif(feat['type'] == bc.FEATURE_ORIT):
                    annotations = {
                        'ID': feat['id'],
                        'Name': feat['product']
                    }
                    if(cfg.compliant):
                        annotations['Note'] = feat['product']
                    else:
                        annotations['product'] = feat['product']
                        annotations['inference'] = 'similar to DNA sequence'
                    annotations = encode_annotations(annotations)
                    feat_type = bc.INSDC_FEATURE_ORIGIN_TRANSFER if cfg.compliant else so.SO_ORIT.name
                    fh.write(f"{seq_id}\tBLAST+\t{feat_type}\t{start}\t{stop}\t.\t{feat['strand']}\t.\t{annotations}\n")

        if(not cfg.compliant):
            fh.write('##FASTA\n')
            for seq in data['sequences']:  # write sequences
                fh.write(f">{seq['id']}\n")
                seq_nt = seq['nt'] if 'nt' in seq else seq['sequence']  # <1.10.0 compatibility
                fh.write(fasta.wrap_sequence(seq_nt))
    return


def encode_attribute(product: str) -> str:
    """Replace special characters forbidden in column 9 of the GFF3 format: https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md"""
    product = str(product)
    product = product.replace('%', '%25')
    product = product.replace(';', '%3B')
    product = product.replace('=', '%3D')
    product = product.replace('&', '%26')
    product = product.replace(',', '%2C')
    return product


def encode_annotations(annotations: Dict[str, Union[str, Sequence[str]]]) -> str:
    annotation_strings = []
    for key, val in annotations.items():
        if(type(val) is list):
            if(len(val) >= 1):
                val = [encode_attribute(k) for k in val]
                annotation = f"{key}={','.join(val)}"
                annotation_strings.append(annotation)
        else:
            annotation_strings.append(f'{key}={encode_attribute(val)}')
    return ';'.join(annotation_strings)


def write_signal_peptide(fh, feat: dict):  # <1.10.0 compatibility
    sig_peptide = feat[bc.FEATURE_SIGNAL_PEPTIDE]
    annotations = {
        'ID': f"{feat['locus']}_sigpep",
        'Name': 'signal peptide',
        'product': 'signal peptide',
        'score': sig_peptide['score'],
        'Parent': feat['locus']
    }
    annotations = encode_annotations(annotations)
    seq_id = feat['sequence'] if 'sequence' in feat else feat['contig']  # <1.10.0 compatibility
    fh.write(f"{seq_id}\tDeepSig\t{so.SO_SIGNAL_PEPTIDE.name}\t{sig_peptide['start']}\t{sig_peptide['stop']}\t{sig_peptide['score']:.2f}\t{feat['strand']}\t.\t{annotations}\n")



import logging
import subprocess as sp

from collections import OrderedDict
from pathlib import Path

import bakta.features.annotation as ba
import bakta.config as cfg
import bakta.constants as bc
import bakta.so as so
import bakta.utils as bu


HIT_EVALUE = 1E-4


log = logging.getLogger('NC_RNA')


def predict_nc_rnas(data: dict, sequences_path: Path):
    """Search for non-coding RNA genes."""

    output_path = cfg.tmp_path.joinpath('ncrna-genes.tsv')
    cmd = [
        'cmscan',
        '--noali',
        '--cut_tc',
        '-g',  # activate glocal mode
        '--nohmmonly',  # strictly use CM models
        '--rfam',
        '--cpu', str(cfg.threads),
        '--tblout', str(output_path)
    ]
    if(data['stats']['size'] >= 1000000):
        cmd.append('-Z')
        cmd.append(str(2 * data['stats']['size'] // 1000000))
    cmd.append(str(cfg.db_path.joinpath('ncRNA-genes')))
    cmd.append(str(sequences_path))
    log.debug('cmd=%s', cmd)
    proc = sp.run(
        cmd,
        cwd=str(cfg.tmp_path),
        env=cfg.env,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        universal_newlines=True
    )
    if(proc.returncode != 0):
        log.debug('stdout=\'%s\', stderr=\'%s\'', proc.stdout, proc.stderr)
        log.warning('ncRNAs failed! cmscan-error-code=%d', proc.returncode)
        raise Exception(f'cmscan error! error code: {proc.returncode}')

    rfam2go = {}
    rfam2go_path = cfg.db_path.joinpath('rfam-go.tsv')
    with rfam2go_path.open() as fh:
        for line in fh:
            (rfam, go) = line.split('\t')
            if(rfam in rfam2go):
                rfam2go[rfam].append(go)
            else:
                rfam2go[rfam] = [go]

    ncrnas = []
    sequences = {seq['id']: seq for seq in data['sequences']}
    with output_path.open() as fh:
        for line in fh:
            if(line[0] != '#'):
                (
                    subject, accession, sequence_id, sequence_acc, mdl, mdl_from, mdl_to,
                    start, stop, strand, trunc, passed, gc, bias, score, evalue,
                    inc, description
                ) = bc.RE_MULTIWHITESPACE.split(line.strip(), maxsplit=17)

                if(strand == '-'):
                    (start, stop) = (stop, start)
                (start, stop) = (int(start), int(stop))
                evalue = float(evalue)
                score = float(score)
                length = stop - start + 1
                if(trunc == "5'"):
                    truncated = bc.FEATURE_END_5_PRIME
                elif(trunc == "3'"):
                    truncated = bc.FEATURE_END_3_PRIME
                else:
                    truncated = None

                if(evalue > HIT_EVALUE):
                    log.debug(
                        'discard low E value: seq=%s, start=%i, stop=%i, strand=%s, gene=%s, length=%i, truncated=%s, score=%1.1f, evalue=%1.1e',
                        sequence_id, start, stop, strand, subject, length, truncated, score, evalue
                    )
                else:
                    rfam_id = f'{bc.DB_XREF_RFAM}:{accession}'
                    db_xrefs = [rfam_id]
                    if(rfam_id in rfam2go):
                        db_xrefs += rfam2go[rfam_id]

                    ncrna = OrderedDict()
                    ncrna['type'] = bc.FEATURE_NC_RNA
                    ncrna['class'] = determine_class(description)
                    ncrna['sequence'] = sequence_id
                    ncrna['start'] = start
                    ncrna['stop'] = stop
                    ncrna['strand'] = bc.STRAND_FORWARD if strand == '+' else bc.STRAND_REVERSE
                    
                    gene = subject
                    if(ba.RE_PROTEIN_SYMBOL.fullmatch(gene)):
                        gene = gene[0].lower() + gene[1:]
                        log.debug('fix gene: lowercase first char. new=%s, old=%s', gene, subject)
                    ncrna['gene'] = gene
                    ncrna['product'] = description

                    if(ncrna['class'] is not None):
                        db_xrefs.append(ncrna['class'].id)
                    else:
                        db_xrefs.append(so.SO_NCRNA_GENE.id)

                    if(truncated):
                        ncrna['truncated'] = truncated

                    ncrna['score'] = score
                    ncrna['evalue'] = evalue
                    ncrna['db_xrefs'] = db_xrefs

                    nt = bu.extract_feature_sequence(ncrna, sequences[sequence_id])  # extract nt sequences
                    ncrna['nt'] = nt

                    ncrnas.append(ncrna)
                    log.info(
                        'seq=%s, start=%i, stop=%i, strand=%s, gene=%s, product=%s, length=%i, truncated=%s, score=%1.1f, evalue=%1.1e, nt=[%s..%s]',
                        ncrna['sequence'], ncrna['start'], ncrna['stop'], ncrna['strand'], ncrna['gene'], ncrna['product'], length, truncated, ncrna['score'], ncrna['evalue'], nt[:10], nt[-10:]
                    )
    log.info('predicted=%i', len(ncrnas))
    return ncrnas


def determine_class(description: str) -> str:
    description = description.lower()
    if('ribozyme' in description):
        return so.SO_NCRNA_GENE_RIBOZYME
    elif('rnase p' in description):
        return so.SO_NCRNA_GENE_RNASEP
    elif('antisense' in description):
        return so.SO_NCRNA_GENE_ANTISENSE
    else:
        None



import atexit
import copy
import json
import logging
import os
import sys

from datetime import datetime
from pathlib import Path

import numpy as np
import yaml

from Bio.SeqFeature import FeatureLocation, CompoundLocation, AfterPosition, BeforePosition
from matplotlib.patches import Patch
from matplotlib.lines import Line2D
from pycirclize import Circos
from xopen import xopen

import bakta
import bakta.utils as bu
import bakta.config as cfg
import bakta.constants as bc
import bakta.io.insdc as insdc


COLORS = {
    'backbone': '#000000',
    'gc-positive': '#33a02c',
    'gc-negative': '#e31a1c',
    'gc-skew-positive': '#fdbf6f',
    'gc-skew-negative': '#1f78b4',
    'features': {  # feature type colors
        bc.FEATURE_CDS: '#cccccc',
        bc.FEATURE_SORF: '#cccccc',
        bc.FEATURE_T_RNA: '#b2df8a',
        bc.FEATURE_TM_RNA: '#b2df8a',
        bc.FEATURE_R_RNA: '#fb8072',
        bc.FEATURE_NC_RNA: '#fdb462',
        bc.FEATURE_NC_RNA_REGION : '#80b1d3',
        bc.FEATURE_CRISPR: '#bebada',
        bc.FEATURE_GAP: '#000000',
        'misc': '#666666'
    },
    'cog-classes': {
        'A': '#c5000b',  # RNA processing and modification
        'B': '#6a3d9a',  # Chromatin structure and dynamics
        'C': '#bebada',  # Energy production and conversion
        'D': '#ff7f00',  # Cell cycle control, cell division, chromosome partitioning
        'E': '#ffffb3',  # Amino acid transport and metabolism
        'F': '#e31a1c',  # Nucleotide transport and metabolism
        'G': '#fb8072',  # Carbohydrate transport and metabolism
        'H': '#b2df8a',  # Coenzyme transport and metabolism
        'I': '#ccebc5',  # Lipid transport and metabolism
        'J': '#a6cee3',  # Translation, ribosomal structure and biogenesis
        'K': '#b3de69',  # Transcription
        'L': '#80b1d3',  # Replication, recombination and repair
        'M': '#bc80bd',  # Cell wall/membrane/envelope biogenesis
        'N': '#cab2d6',  # Cell motility
        'O': '#33a02c',  # Posttranslational modification, protein turnover, chaperones
        'P': '#fccde5',  # Inorganic ion transport and metabolism
        'Q': '#1f78b4',  # Secondary metabolites biosynthesis, transport and catabolism
        'R': '#8dd3c7',  # General function prediction only
        'S': '#222222',  # Function unknown
        'T': '#fdb462',  # Signal transduction mechanisms
        'U': '#fdbf6f',  # Intracellular trafficking, secretion, and vesicular transport
        'V': '#fb9a99',  # Defense mechanisms
        'W': '#0084d1',  # Extracellular structures
        'X': '#d9d9d9',  # Mobilome: prophages, transposons
        'Y': '#ffff38',  # Nuclear structure
        'Z': '#ffff99'   # Cytoskeleton
    }
}


log = logging.getLogger('PLOT')


def main():
    # parse options and arguments
    parser = bu.init_parser(sub_command='_plot')
    parser.add_argument('input', metavar='<input>', help='Bakta annotations in JSON format')
    
    arg_group_io = parser.add_argument_group('Input / Output')
    arg_group_io.add_argument('--config', '-c', action='store', default=None, help='Plotting configuration in YAML format')
    arg_group_io.add_argument('--output', '-o', action='store', default=os.getcwd(), help='Output directory (default = current working directory)')
    arg_group_io.add_argument('--prefix', '-p', action='store', default=None, help='Prefix for output files')
    arg_group_io.add_argument('--force', '-f', action='store_true', help='Force overwriting existing output folder')

    arg_group_plot = parser.add_argument_group('Plotting')
    arg_group_plot.add_argument('--sequences', action='store', default='all', help='Sequences to plot: comma separated number or name (default = all, numbers one-based)')
    arg_group_plot.add_argument('--type', action='store', type=str, default=bc.PLOT_FEATURES, choices=[bc.PLOT_FEATURES, bc.PLOT_COG], help=f'Plot type (default = {bc.PLOT_FEATURES})')
    arg_group_plot.add_argument('--label', action='store', type=str, default=None, help=f"Plot center label (for line breaks use '|')")
    arg_group_plot.add_argument('--size', action='store', type=int, default=8, choices=[4, 8, 16], help='Plot size in inches: 4/8/16 (default = 8)')
    arg_group_plot.add_argument('--dpi', action='store', type=int, default=300, choices=[150, 300, 600], help='Plot resolution as dots per inch: 150/300/600 (default = 300)')

    arg_group_general = parser.add_argument_group('General')
    arg_group_general.add_argument('--help', '-h', action='help', help='Show this help message and exit')
    arg_group_general.add_argument('--verbose', '-v', action='store_true', help='Print verbose information')
    arg_group_general.add_argument('--debug', action='store_true', help='Run Bakta in debug mode. Temp data will not be removed.')
    arg_group_general.add_argument('--tmp-dir', action='store', default=None, dest='tmp_dir', help='Location for temporary files (default = system dependent auto detection)')
    arg_group_general.add_argument('--version', action='version', version=f'%(prog)s {cfg.version}')
    args = parser.parse_args()

    ############################################################################
    # Setup logging
    ############################################################################
    cfg.prefix = args.prefix if args.prefix else Path(args.input).stem
    output_path = cfg.check_output_path(args.output, args.force)
    cfg.force = args.force
    log.info('force=%s', args.force)
    
    bu.setup_logger(output_path, cfg.prefix, args)
    log.info('prefix=%s', cfg.prefix)
    log.info('output=%s', output_path)

    
    ############################################################################
    # Checks and configurations
    ############################################################################
    try:
        if args.input == '':
            raise ValueError('File path argument must be non-empty')
        annotation_path = Path(args.input).resolve()
        cfg.check_readability('annotation', annotation_path)
        cfg.check_content_size('annotation', annotation_path)
    except:
        log.error('provided annotation file not valid! path=%s', args.input)
        sys.exit(f'ERROR: annotation file ({args.input}) not valid!')
    
    log.info('input-path=%s', annotation_path)
    cfg.check_tmp_path(args)
    cfg.debug = args.debug
    log.info('debug=%s', cfg.debug)
    cfg.verbose = True if cfg.debug else args.verbose
    log.info('verbose=%s', cfg.verbose)
    plot_type = args.type
    log.info('plot-type=%s', plot_type)
    plot_label = args.label
    log.info('plot-label=%s', plot_label)
    plot_size = args.size
    log.info('plot-size=%s', plot_size)
    plot_dpi = args.dpi
    log.info('plot-dpi=%s', plot_dpi)

    # check and open configuration file
    if args.config is not None:
        config_path = None
        try:
            if(args.config == ''):
                raise ValueError('File path argument must be non-empty')
            config_path = Path(args.config).resolve()
            cfg.check_readability('config', config_path)
            cfg.check_content_size('config', config_path)
            log.info('config=%s', config_path)
            with config_path.open() as fh:
                config = yaml.load(fh, Loader=yaml.FullLoader)
        except:
            log.error('provided config file not valid! path=%s', args.config)
            sys.exit(f'ERROR: config file ({args.config}) not valid!')
    else:
        config = {}

    bu.test_dependencies()
    if(cfg.verbose):
        print(f'Bakta v{cfg.version}')
        print('Options and arguments:')
        print(f'\tinput: {annotation_path}')
        if(args.config): print(f'\tconfig: {args.config}')
        print(f'\toutput: {cfg.output_path}')
        if(cfg.force): print(f'\tforce: {cfg.force}')
        print(f'\ttmp directory: {cfg.tmp_path}')
        print(f'\tprefix: {cfg.prefix}')
        print(f'\tlabel: {plot_label}')
        print(f'\tsize: {plot_size}')
        print(f'\tDPI: {plot_dpi}')

    if(cfg.debug):
        print(f"\nBakta runs in DEBUG mode! Temporary data will not be destroyed at: {cfg.tmp_path}")
    else:
        atexit.register(bu.cleanup, log, cfg.tmp_path)  # register cleanup exit hook

    ############################################################################
    # Write plots
    ############################################################################

    # load genome annotations
    print('Parse genome annotations...')
    with xopen(str(annotation_path), threads=0) as fh:
        data = json.load(fh)
    
    # set global config objects based on information from imported JSON document
    cfg.db_info = {
        'type': data['version']['db']['type'],
        'major': data['version']['db']['version'].split('.')[0],
        'minor': data['version']['db']['version'].split('.')[1]
    }
    cfg.translation_table = data['genome']['translation_table']
    cfg.run_start = datetime.strptime(data['run']['start'], '%Y-%m-%d %H:%M:%S')
    cfg.run_end = datetime.strptime(data['run']['end'], '%Y-%m-%d %H:%M:%S')

    features = data['features']
    sequences = data['sequences']

    # load colors if specified
    colors = COLORS
    conf_colors = config.get('colors', None)
    if conf_colors is not None:
        colors = {**colors, **conf_colors}

    print('Draw plots...')
    if args.sequences == 'all':  # write whole genome plot
        print(f'\tdraw circular genome plot (type={plot_type}) containing all sequences...')
        write(data, features, output_path, colors, plot_type=plot_type, plot_label=plot_label, plot_size=plot_size, plot_dpi=plot_dpi)
    else:  # write genome plot containing provided sequences only
        plot_sequences = []
        sequence_identifiers = []
        for selected_sequence in args.sequences.split(','):
            for i, seq in enumerate(sequences):
                sequence_no = str(i + 1)
                if selected_sequence == sequence_no:
                    plot_sequences.append(seq)
                    sequence_identifiers.append(sequence_no)
                elif selected_sequence.lower() == seq['id'].lower():
                    plot_sequences.append(seq)
                    sequence_identifiers.append(seq['id'])
        if len(plot_sequences) > 0:
            print(f'\tdraw circular genome plot (type={plot_type}) containing sequences: {sequence_identifiers}...')
            plot_name_suffix = '_'.join(sequence_identifiers)
            plot_sequence_ids = [seq['id'] for seq in plot_sequences]
            data['features'] = [feat for feat in features if feat['sequence'] in plot_sequence_ids]  # reduce feature list in data object
            data['sequences'] = [seq for seq in sequences if seq['id'] in plot_sequence_ids]  # reduce sequence list in data object
            write(data, features, output_path, colors, plot_name_suffix=plot_name_suffix, plot_type=plot_type, plot_label=plot_label, plot_size=plot_size, plot_dpi=plot_dpi)


def write(data, features, output_path, colors=COLORS, plot_name_suffix=None, plot_type=bc.PLOT_FEATURES, plot_label=None, plot_size=8, plot_dpi=300):
    sequence_list = insdc.build_biopython_sequence_list(data, features)
    clipped = False
    if(len(sequence_list) > 20):
        sequence_list = sorted(sequence_list, key=lambda a: len(a.seq), reverse=True)[:20]  # select longest 20 sequences in draft mode
        clipped = True
    for seq in sequence_list:  # fix edge features because PyCirclize cannot handle them correctly
        seq.features = [feat for feat in seq.features if feat.type != 'gene' and feat.type != 'source']
        for feat in seq.features:
            if isinstance(feat.location, CompoundLocation):
                feat_loc = feat.location
                log.debug('split edge feature: seq=%s, start=%i, stop=%i, strand=%s', seq.id, feat_loc.start, feat_loc.end, '+' if feat_loc.strand==1 else '-')
                if(feat_loc.strand == +1):
                    feat.location = FeatureLocation(feat_loc.parts[0].start, len(seq.seq), strand=feat_loc.strand)
                    feat_2 = copy.deepcopy(feat)
                    feat_2.location = FeatureLocation(0, feat_loc.parts[1].end, strand=feat_loc.strand)
                    seq.features.append(feat_2)
                elif(feat_loc.strand == -1):
                    feat.location = FeatureLocation(0, feat_loc.parts[0].end, strand=feat_loc.strand)
                    feat_2 = copy.deepcopy(feat)
                    feat_2.location = FeatureLocation(feat_loc.parts[1].start,  len(seq.seq), strand=feat_loc.strand)
                    seq.features.append(feat_2)
            else:
                if isinstance(feat.location.start, AfterPosition) or isinstance(feat.location.start, BeforePosition):
                    feat.location = FeatureLocation(int(str(feat.location.start)[1:]), feat.location.end, strand=feat.location.strand)
                if isinstance(feat.location.end, AfterPosition) or isinstance(feat.location.end, BeforePosition):
                    feat.location = FeatureLocation(feat.location.start, int(str(feat.location.end)[1:]), strand=feat.location.strand)

    # build lable
    plot_label = build_label(data) if plot_label is None else plot_label.replace('|', '\n')

    # select style
    if plot_type == bc.PLOT_COG:
        plot = build_features_type_cog(data, sequence_list, clipped, colors, plot_label, plot_size, plot_dpi)
    else:
        plot = build_features_type_feature(data, sequence_list, clipped, colors, plot_label, plot_size, plot_dpi)
    file_name = cfg.prefix if plot_name_suffix is None else f'{cfg.prefix}_{plot_name_suffix}'
    for file_type in ['png', 'svg']:
        file_path = output_path.joinpath(f'{file_name}.{file_type}')
        plot.savefig(file_path)


def build_label(data):
    genus = data['genome'].get('genus', '') if data['genome'].get('genus', None) else ''
    species = data['genome'].get('species', '') if data['genome'].get('species', None) else ''
    taxon = ' '.join([genus, species]).replace('  ', ' ')
    if taxon == '' or taxon == ' ':
        taxon = None
    strain = data['genome'].get('strain', '') if data['genome'].get('strain', None) else ''
    genome_size = sum([len(seq['nt']) for seq in data['sequences']]) if 'nt' in data['sequences'][0] else sum([len(seq['sequence']) for seq in data['sequences']])  # <1.10.0 compatibility
    genome_size_lable = f'{(genome_size/(10**6)):0.1f} Mbp' if genome_size >= 10**6 else f'{(genome_size/(10**3)):0.1f} kbp'
    label_list = []
    sequence_number = len(data['sequences'])
    if sequence_number == 1:
        seq_type = bc.REPLICON_CONTIG
        for seq in data['sequences']:
            if seq['id'] == data['sequences'][0]['id']:
                seq_type = seq['type']
        if seq_type == bc.REPLICON_CHROMOSOME:
            label_list.append(taxon)
            label_list.append(strain)
            label_list.append(f'chromosome, {genome_size_lable}')
        elif seq_type == bc.REPLICON_PLASMID:
            plasmid_name = data['genome'].get('plasmid', None)
            label_list.append(plasmid_name if plasmid_name else seq['id'])
            label_list.append(f'plasmid, {genome_size_lable}')
        else:  # a single contig
            label_list.append(taxon)
            label_list.append(strain)
            label_list.append(f"{seq['id']}, {genome_size_lable}")
    else:  # full (draft) genome
        label_list.append(taxon)
        label_list.append(strain)
        label_list.append(f'{sequence_number} sequences, {genome_size_lable}')
    return '\n'.join([lable for lable in label_list if lable is not None])


def build_features_type_feature(data, sequence_list, clipped, colors, plot_label, plot_size, plot_dpi):
    # Get contig genome seqid & size, features dict
    total_sequence_length = sum([len(seq['nt']) for seq in data['sequences']]) if 'nt' in data['sequences'][0] else sum([len(seq['sequence']) for seq in data['sequences']])  # <1.10.0 compatibility
    seqid2seq = {rec.id:rec.seq for rec in sequence_list}
    seqid2size = {rec.id:len(rec.seq) for rec in sequence_list}
    seqid2features = {rec.id:rec.features for rec in sequence_list}

    if plot_size == 4:
        text_size = 6
    elif plot_size == 8:
        text_size = 12
    else:
        text_size = 30
    
    if(clipped):
        circos = Circos(seqid2size, space=2, start=10, end=350)
        circos.text('...', r=99.5, color=colors['backbone'], ha='center', va='center', size=10)
    else:
        circos = Circos(seqid2size, space=2)
    circos.text(plot_label, r=7, size=text_size, linespacing=1.5)
    for sector in circos.sectors:
        # build tracks
        outer_track = sector.add_track((99.5, 100))
        feature_forward_track = sector.add_track((87, 97), r_pad_ratio=0.1)
        feature_reverse_track = sector.add_track((77, 87), r_pad_ratio=0.1)
        gc_content_track = sector.add_track((62, 72))
        gc_skew_track = sector.add_track((50, 60))

        # plot outer track
        build_sequence_backbone_track(sector, outer_track, total_sequence_length, colors, plot_size)

        # plot feature tracks
        for feature in seqid2features[sector.name]:
            track = feature_forward_track if feature.location.strand == 1 else feature_reverse_track
            if feature.type == bc.INSDC_FEATURE_CDS:
                track.genomic_features([feature], fc=colors['features'][bc.FEATURE_CDS])
            elif feature.type == bc.INSDC_FEATURE_T_RNA:
                track.genomic_features([feature], fc=colors['features'][bc.FEATURE_T_RNA])
            elif feature.type == bc.INSDC_FEATURE_TM_RNA:
                track.genomic_features([feature], fc=colors['features'][bc.FEATURE_TM_RNA])
            elif feature.type == bc.INSDC_FEATURE_R_RNA:
                track.genomic_features([feature], fc=colors['features'][bc.FEATURE_R_RNA])
            elif feature.type == bc.INSDC_FEATURE_NC_RNA:
                track.genomic_features([feature], fc=colors['features'][bc.FEATURE_NC_RNA])
            elif feature.type == bc.INSDC_FEATURE_REGULATORY:
                track.genomic_features([feature], fc=colors['features'][bc.FEATURE_NC_RNA_REGION])
            elif feature.type == bc.INSDC_FEATURE_REPEAT_REGION:
                track.genomic_features([feature], fc=colors['features'][bc.FEATURE_CRISPR])
            elif feature.type == bc.INSDC_FEATURE_GAP:
                track.genomic_features([feature], fc=colors['features'][bc.FEATURE_GAP])
            elif feature.type == bc.INSDC_FEATURE_ORIGIN_REPLICATION:
                gc_skew_track.xticks([(feature.location.start + feature.location.end)/2], outer=False, label_size=text_size/2, labels=['oriC'], label_orientation='vertical')  # oriC/V
            elif feature.type == bc.INSDC_FEATURE_ORIGIN_TRANSFER:
                gc_skew_track.xticks([(feature.location.start + feature.location.end)/2], outer=False, label_size=text_size/2, labels=['oriT'], label_orientation='vertical')  # oriT
            else:
                track.genomic_features([feature], fc=colors['features']['misc'])
    
        # plot GC content and GC skew
        seq = str(seqid2seq[sector.name])
        build_gc_content_skew(seq, colors, gc_content_track, gc_skew_track)

    fig = circos.plotfig(dpi=plot_dpi, figsize=(plot_size,plot_size))
    build_legend(circos, colors, plot_size)
    return fig


def build_features_type_cog(data, sequence_list, clipped, colors, plot_label, plot_size, plot_dpi):
    # Get contig genome seqid & size, features dict
    total_sequence_length = sum([len(seq['nt']) for seq in data['sequences']])
    seqid2seq = {rec.id:rec.seq for rec in sequence_list}
    seqid2size = {rec.id:len(rec.seq) for rec in sequence_list}
    seqid2features = {rec.id:rec.features for rec in sequence_list}

    if plot_size == 4:
        text_size = 6
    elif plot_size == 8:
        text_size = 12
    else:
        text_size = 30
    
    if(clipped):
        circos = Circos(seqid2size, space=2, start=10, end=350)
        circos.text('...', r=99.5, color=colors['backbone'], ha='center', va='center', size=6)
    else:
        circos = Circos(seqid2size, space=2)
    circos.text(plot_label, r=7, size=text_size, linespacing=1.5)
    for sector in circos.sectors:
        # build tracks
        outer_track = sector.add_track((99.5, 100))
        feature_forward_track = sector.add_track((87, 97), r_pad_ratio=0.1)
        feature_reverse_track = sector.add_track((77, 87), r_pad_ratio=0.1)
        non_cds_feature_track = sector.add_track((67, 77), r_pad_ratio=0.1)
        gc_content_track = sector.add_track((57, 67))
        gc_skew_track = sector.add_track((45, 55))

        # plot outer track
        build_sequence_backbone_track(sector, outer_track, total_sequence_length, colors, plot_size)

        # plot feature tracks
        for feature in seqid2features[sector.name]:
            if feature.type == bc.INSDC_FEATURE_CDS:
                color = colors['features'][bc.FEATURE_CDS]
                dbxrefs = feature.qualifiers.get('db_xref', None)
                if dbxrefs is not None:
                    for dbxref in dbxrefs:
                        if bc.DB_PREFIX_COG in dbxref:
                            cog = dbxref.split(':')[1]
                            if len(cog) != 7:  # skip COG clusters
                                if len(cog) != 1:  # select first of two COG categories
                                    cog = cog[:1]
                                color = colors['cog-classes'].get(cog.upper(), colors['cog-classes']['S'])
                track = feature_forward_track if feature.location.strand == 1 else feature_reverse_track
                track.genomic_features([feature], fc=color)
            elif feature.type == bc.INSDC_FEATURE_T_RNA:
                non_cds_feature_track.genomic_features([feature], fc=colors['features'][bc.FEATURE_T_RNA])
            elif feature.type == bc.INSDC_FEATURE_TM_RNA:
                non_cds_feature_track.genomic_features([feature], fc=colors['features'][bc.FEATURE_TM_RNA])
            elif feature.type == bc.INSDC_FEATURE_R_RNA:
                non_cds_feature_track.genomic_features([feature], fc=colors['features'][bc.FEATURE_R_RNA])
            elif feature.type == bc.INSDC_FEATURE_NC_RNA:
                non_cds_feature_track.genomic_features([feature], fc=colors['features'][bc.FEATURE_NC_RNA])
            elif feature.type == bc.INSDC_FEATURE_REGULATORY:
                non_cds_feature_track.genomic_features([feature], fc=colors['features'][bc.FEATURE_NC_RNA_REGION])
            elif feature.type == bc.INSDC_FEATURE_REPEAT_REGION:
                non_cds_feature_track.genomic_features([feature], fc=colors['features'][bc.FEATURE_CRISPR])
            elif feature.type == bc.INSDC_FEATURE_GAP:
                non_cds_feature_track.genomic_features([feature], fc=colors['features'][bc.FEATURE_GAP])
            elif feature.type == bc.INSDC_FEATURE_ORIGIN_REPLICATION:
                gc_skew_track.xticks([(feature.location.start + feature.location.end)/2], outer=False, label_size=text_size/2, labels=['oriC'], label_orientation='vertical')  # oriC/V
            elif feature.type == bc.INSDC_FEATURE_ORIGIN_TRANSFER:
                gc_skew_track.xticks([(feature.location.start + feature.location.end)/2], outer=False, label_size=text_size/2, labels=['oriT'], label_orientation='vertical')  # oriT
            else:
                non_cds_feature_track.genomic_features([feature], fc=colors['features']['misc'])
    
        # plot GC content and GC skew
        seq = str(seqid2seq[sector.name])
        build_gc_content_skew(seq, colors, gc_content_track, gc_skew_track)

    fig = circos.plotfig(dpi=plot_dpi, figsize=(plot_size,plot_size))
    build_legend(circos, colors, plot_size)
    return fig


def build_sequence_backbone_track(sector, outer_track, total_sequence_length, colors, plot_size):
    outer_track.axis(fc=colors['backbone'])
    label_formatter=lambda v: f'{v / 1000:.1f} kbp'
    if total_sequence_length >= 1_000_000:
        major_interval = 500_000
        minor_interval = int(major_interval / 5)
        label_formatter=lambda v: f'{v / 1_000_000:.1f} Mbp'
    elif total_sequence_length >= 100_000:
        major_interval = 100_000
        minor_interval = int(major_interval / 10)
    elif total_sequence_length >= 10_000:
        major_interval = 10_000
        minor_interval = int(major_interval / 10)
    elif total_sequence_length >= 1_000:
        major_interval = 1_000
        minor_interval = int(major_interval / 10)
    else:
        major_interval = 100
        minor_interval = int(major_interval / 10)
    
    if plot_size == 4:
        text_size = 4
    elif plot_size == 8:
        text_size = 8
    else:
        text_size = 16
    
    if sector.size > minor_interval:
        outer_track.xticks_by_interval(major_interval, label_formatter=label_formatter, label_size=text_size)
        outer_track.xticks_by_interval(minor_interval, tick_length=1, show_label=False, label_size=text_size)


def build_gc_content_skew(sequence, colors, gc_content_track, gc_skew_track):
    pos_list, gc_contents = calc_gc_content(sequence)
    gc_contents = gc_contents - gc_fraction(sequence) * 100
    positive_gc_contents = np.where(gc_contents > 0, gc_contents, 0)
    negative_gc_contents = np.where(gc_contents < 0, gc_contents, 0)
    abs_max_gc_content = np.max(np.abs(gc_contents))
    gc_content_track.fill_between(pos_list, positive_gc_contents, 0, vmin=-abs_max_gc_content, vmax=abs_max_gc_content, color=colors['gc-positive'])
    gc_content_track.fill_between(pos_list, negative_gc_contents, 0, vmin=-abs_max_gc_content, vmax=abs_max_gc_content, color=colors['gc-negative'])

    pos_list, gc_skews = calc_gc_skew(sequence)
    positive_gc_skews = np.where(gc_skews > 0, gc_skews, 0)
    negative_gc_skews = np.where(gc_skews < 0, gc_skews, 0)
    abs_max_gc_skew = np.max(np.abs(gc_skews))
    gc_skew_track.fill_between(pos_list, positive_gc_skews, 0, vmin=-abs_max_gc_skew, vmax=abs_max_gc_skew, color=colors['gc-skew-positive'])
    gc_skew_track.fill_between(pos_list, negative_gc_skews, 0, vmin=-abs_max_gc_skew, vmax=abs_max_gc_skew, color=colors['gc-skew-negative'])


def build_legend(circos, colors, plot_size):

    if plot_size == 4:
        text_size = 3
        marker_size = 1
    elif plot_size == 8:
        text_size = 6
        marker_size = 5
    else:
        text_size = 12
        marker_size = 9

    handles=[
        Patch(color=colors['features'][bc.FEATURE_CDS], label='CDS'),
        Patch(color=colors['features'][bc.FEATURE_T_RNA], label='tRNA'),
        Patch(color=colors['features'][bc.FEATURE_R_RNA], label='rRNA'),
        Patch(color=colors['features'][bc.FEATURE_NC_RNA], label='ncRNA'),
        Patch(color=colors['features'][bc.FEATURE_NC_RNA_REGION], label='ncRNA reg'),
        Patch(color=colors['features'][bc.FEATURE_CRISPR], label='CRISPR'),
        Line2D([], [], color=colors['gc-positive'], label="+ GC", marker="^", ms=marker_size, ls="None"),
        Line2D([], [], color=colors['gc-negative'], label="- GC", marker="v", ms=marker_size, ls="None"),
        Line2D([], [], color=colors['gc-skew-positive'], label="+ GC Skew", marker="^", ms=marker_size, ls="None"),
        Line2D([], [], color=colors['gc-skew-negative'], label="- GC Skew", marker="v", ms=marker_size, ls="None")
    ]
    _ = circos.ax.legend(
        handles=handles,
        bbox_to_anchor=(0.5, 0.4),
        loc='center',
        ncols=2,
        fontsize=text_size
    )


def calc_gc_content(seq: str):
    pos_list, gc_content_list = [], []
    window_size = int(len(seq) / 500)
    step_size = int(len(seq) / 1000)
    if window_size == 0 or step_size == 0:
        window_size, step_size = len(seq), int(len(seq) / 2)
    pos_list = list(range(0, len(seq), step_size)) + [len(seq)]
    for pos in pos_list:
        window_start_pos = pos - int(window_size / 2)
        window_end_pos = pos + int(window_size / 2)
        if window_start_pos < 0:
            window_start_pos = 0
        if window_end_pos > len(seq):
            window_end_pos = len(seq)
        subseq = seq[window_start_pos:window_end_pos]
        gc_content = gc_fraction(subseq) * 100
        gc_content_list.append(gc_content)
    return np.array(pos_list).astype(np.int64), np.array(gc_content_list).astype(np.float64)


def calc_gc_skew(seq: str):
    pos_list, gc_skew_list = [], []
    window_size = int(len(seq) / 500)
    step_size = int(len(seq) / 1000)
    if window_size == 0 or step_size == 0:
        window_size, step_size = len(seq), int(len(seq) / 2)
    pos_list = list(range(0, len(seq), step_size)) + [len(seq)]
    for pos in pos_list:
        window_start_pos = pos - int(window_size / 2)
        window_end_pos = pos + int(window_size / 2)
        window_start_pos = 0 if window_start_pos < 0 else window_start_pos
        window_end_pos = len(seq) if window_end_pos > len(seq) else window_end_pos
        subseq = seq[window_start_pos:window_end_pos]
        g, c = subseq.count('G'), subseq.count('C')
        gc_skew = (g - c) / float(g + c) if (g + c) > 0 else 0.0
        gc_skew_list.append(gc_skew)
    return np.array(pos_list).astype(np.int64), np.array(gc_skew_list).astype(np.float64)


def gc_fraction(seq: str):
    gc = seq.count('C') + seq.count('G') + seq.count('S')
    length = len(seq)
    return 0 if length == 0 else gc / length

if __name__ == '__main__':
    main()




import logging
import re

from typing import Sequence

import bakta.config as cfg
import bakta.constants as bc
import bakta.io.insdc as insdc


log = logging.getLogger('ANNOTATION')


RE_MULTIWHITESPACE = re.compile(r'\s{2,}')

RE_PROTEIN_CONTIG = re.compile(r'\s+contig\s*', flags=re.IGNORECASE)
RE_PROTEIN_HOMOLOG = re.compile(r'\shomolog(?: (\d+))?', flags=re.IGNORECASE)
RE_PROTEIN_PUTATIVE = re.compile(r'(potential|possible|probable|predicted)', flags=re.IGNORECASE)
RE_PROTEIN_NODE = re.compile(r'NODE_', flags=re.IGNORECASE)
RE_PROTEIN_POTENTIAL_CONTIG_NAME = re.compile(r'(genome|shotgun)', flags=re.IGNORECASE)
RE_PROTEIN_DOMAIN_CONTAINING = re.compile(r'domain-containing protein', flags=re.IGNORECASE)
RE_PROTEIN_REMNANT = re.compile(r'Remnant of ', re.IGNORECASE)
RE_PROTEIN_TMRNA = re.compile(r'TmRNA', flags=re.IGNORECASE)
RE_PROTEIN_NO_LETTERS = re.compile(r'[^A-Za-z]')
RE_PROTEIN_SUSPECT_CHARS_DISCARD = re.compile(r'[.#]')
RE_PROTEIN_SUSPECT_CHARS_REPLACE = re.compile(r'[@=?%]')
RE_PROTEIN_SUSPECT_CHARS_BEGINNING = '_\-+.:,;/\\\''
RE_PROTEIN_PERIOD_SEPARATOR = re.compile(r'([a-zA-Z0-9]+)\.([a-zA-Z0-9]+)')
RE_PROTEIN_WRONG_PRIMES = re.compile(r'[\u2032\u0060\u00B4]')  # prime (′), grave accent (`), acute accent (´)
RE_PROTEIN_WEIGHT = re.compile(r' [0-9]+(?:\.[0-9]+)? k?da ', flags=re.IGNORECASE)
RE_PROTEIN_SYMBOL = re.compile(r'[A-Z][a-z]{2}[A-Z][0-9]?')
RE_DOMAIN_OF_UNKNOWN_FUNCTION = re.compile(r'(DUF\d{3,4})', flags=re.IGNORECASE)
RE_UNCHARACTERIZED_PROTEIN_FAMILY = re.compile(r'(UPF\d{3,4})', flags=re.IGNORECASE)
RE_GENE_CAPITALIZED = re.compile(r'^[A-Z].+', flags=re.DOTALL)
RE_GENE_SUSPECT_CHARS = re.compile(r'[\?]', flags=re.DOTALL)
RE_GENE_SYMBOL = re.compile(r'[a-z]{3}[A-Z][0-9]?')


def combine_annotation(feature: dict):
    pseudogene_psc = None
    pseudogene_pscc = None
    if(bc.PSEUDOGENE in feature):
        pseudogene_psc = feature[bc.PSEUDOGENE].get('psc', None)
        pseudogene_pscc = feature[bc.PSEUDOGENE].get('pscc', None)
    ups = feature.get('ups', None)
    ips = feature.get('ips', None)
    psc = feature.get('psc', None)
    pscc = feature.get('pscc', None)
    expert_hits = feature.get('expert', [])

    gene = None
    genes = set()
    product = None
    db_xrefs = set()
    if(pseudogene_pscc):
        pseudogene_pscc_genes = pseudogene_pscc.get('gene', None)
        if(pseudogene_pscc_genes):
            pseudogene_pscc_genes = pseudogene_pscc_genes.replace('/', ',').split(',')
            genes.update(pseudogene_pscc_genes)
            gene = pseudogene_pscc_genes[0]
        pscc_product = pseudogene_pscc.get('product', None)
        if(pscc_product):
            product = pscc_product
        for db_xref in pseudogene_pscc['db_xrefs']:
            db_xrefs.add(db_xref)
    if(pseudogene_psc):
        pseudogene_psc_genes = pseudogene_psc.get('gene', None)
        if(pseudogene_psc_genes):
            pseudogene_psc_genes = pseudogene_psc_genes.replace('/', ',').split(',')
            genes.update(pseudogene_psc_genes)
            gene = pseudogene_psc_genes[0]
        psc_product = pseudogene_psc.get('product', None)
        if(psc_product):
            product = psc_product
        for db_xref in pseudogene_psc['db_xrefs']:
            db_xrefs.add(db_xref)
    if(pscc):
        pscc_product = pscc.get('product', None)
        if(pscc_product):
            product = pscc_product
        for db_xref in pscc['db_xrefs']:
            db_xrefs.add(db_xref)
    if(psc and psc.get('valid', True)):
        psc_genes = psc.get('gene', None)
        if(psc_genes):
            psc_genes = psc_genes.replace('/', ',').split(',')
            genes.update(psc_genes)
            gene = psc_genes[0]
        psc_product = psc.get('product', None)
        if(psc_product):
            product = psc_product
        for db_xref in psc['db_xrefs']:
            db_xrefs.add(db_xref)
    if(ups):
        for db_xref in ups['db_xrefs']:
            db_xrefs.add(db_xref)
    if(ips):
        ips_genes = ips.get('gene', None)
        if(ips_genes):
            ips_genes = ips_genes.replace('/', ',').split(',')
            genes.update(ips_genes)
            gene = ips_genes[0]
        ips_product = ips.get('product', None)
        if(ips_product):
            product = ips_product
        for db_xref in ips['db_xrefs']:
            db_xrefs.add(db_xref)

    if(len(expert_hits) > 0):
        top_expert_hit = sorted(expert_hits,key=lambda k: (k['rank'], k.get('score', 0), calc_annotation_score(k)), reverse=True)[0]
        expert_genes = top_expert_hit.get('gene', None)
        if(expert_genes):
            expert_genes = expert_genes.replace('/', ',').split(',')
            genes.update(expert_genes)
            gene = expert_genes[0]
        product = top_expert_hit.get('product', None)
        for hit in expert_hits:
            db_xrefs.update(hit.get('db_xrefs', []))

    if(product):
        product = revise_cds_product(product)
        if(product):
            if(cfg.compliant):
                product = insdc.revise_product_insdc(product)
            feature['product'] = product
        
            protein_gene_symbol = extract_protein_gene_symbol(product)
            if(protein_gene_symbol):
                genes.add(protein_gene_symbol)
            revised_genes = revise_cds_gene_symbols(genes)
            revised_gene = None
            if gene is not None:
                revised_gene = revise_cds_gene_symbols([gene])  # special treatment for selected gene symbol
                revised_gene = revised_gene[0] if len(revised_gene) > 0 else None
            if(revised_gene is None  and  len(revised_genes) >= 1):  # select first from gene symbol list if no symbol was selected before
                revised_gene = revised_genes[0]

            feature['gene'] = revised_gene
            feature['genes'] = sorted(revised_genes)
        else:
            mark_as_hypothetical(feature)
    else:
        mark_as_hypothetical(feature)

    feature['db_xrefs'] = sorted(list(db_xrefs))


def detect_feature_overlaps(data: dict):
    """Apply feature type specific hierarchical feature overlap filters.
    tRNA < tmRNA
    CDS < tmRNA, tRNA, rRNA, CRISPR
    sORF < mRNA, tRNA, rRNA, CRISPR, CDS (in-frame & entirely overlapping), sORF (shorter, weaker annotations)
    """
    sequence_t_rnas = {seq['id']: [] for seq in data['sequences']}
    for trna in [feat for feat in data['features'] if feat['type'] == bc.FEATURE_T_RNA]:
        t_rnas = sequence_t_rnas[trna['sequence']]
        t_rnas.append(trna)
    sequence_tm_rnas = {seq['id']: [] for seq in data['sequences']}
    for tm_rna in [feat for feat in data['features'] if feat['type'] == bc.FEATURE_TM_RNA]:
        tm_rnas = sequence_tm_rnas[tm_rna['sequence']]
        tm_rnas.append(tm_rna)
    sequence_r_rnas = {seq['id']: [] for seq in data['sequences']}
    for r_rna in [feat for feat in data['features'] if feat['type'] == bc.FEATURE_R_RNA]:
        r_rnas = sequence_r_rnas[r_rna['sequence']]
        r_rnas.append(r_rna)
    sequence_ncrna_regions = {seq['id']: [] for seq in data['sequences']}
    for ncRNA_region in [feat for feat in data['features'] if feat['type'] == bc.FEATURE_NC_RNA_REGION]:
        ncRNA_regions = sequence_ncrna_regions[ncRNA_region['sequence']]
        ncRNA_regions.append(ncRNA_region)
    sequence_crispr_arrays = {seq['id']: [] for seq in data['sequences']}
    for crispr_array in [feat for feat in data['features'] if feat['type'] == bc.FEATURE_CRISPR]:
        crispr_arrays = sequence_crispr_arrays[crispr_array['sequence']]
        crispr_arrays.append(crispr_array)
    sequence_cdss = {seq['id']: [] for seq in data['sequences']}
    sequence_cdss_user_provided = {seq['id']: [] for seq in data['sequences']}
    for cds in [feat for feat in data['features'] if feat['type'] == bc.FEATURE_CDS]:
        if(cds.get('source', None) == bc.CDS_SOURCE_USER):
            cdss = sequence_cdss_user_provided[cds['sequence']]
        else:
            cdss = sequence_cdss[cds['sequence']]
        cdss.append(cds)
    sequence_sorfs = {seq['id']: [] for seq in data['sequences']}
    for sorf in [feat for feat in data['features'] if feat['type'] == bc.FEATURE_SORF]:
        sorfs = sequence_sorfs[sorf['sequence']]
        sorfs.append(sorf)

    for seq in data['sequences']:  # find feature overlaps sequence-wise to increase the performance
        log.debug('filter features on seq: %s', seq['id'])

        # mark tRNAs overlapping with tmRNAs
        for tRNA in sequence_t_rnas[seq['id']]:
            for tmRNA in sequence_tm_rnas[seq['id']]:
                if(tRNA['stop'] < tmRNA['start'] or tRNA['start'] > tmRNA['stop']):
                    continue
                else:  # overlap -> remove tRNA
                    overlap = f"[{max(tRNA['start'], tmRNA['start'])},{min(tRNA['stop'], tmRNA['stop'])}]"
                    tRNA['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_TM_RNA,
                        'description': f"{bc.FEATURE_TM_RNA} overlap with ({tmRNA['product']}) at {overlap}"
                    }
                    log.info(
                        "overlap: tRNA (%s) [%i, %i] overlapping with tmRNA (%s) [%i, %i] at %s on seq=%s",
                        tRNA['product'], tRNA['start'], tRNA['stop'], tmRNA['product'], tmRNA['start'], tmRNA['stop'], overlap, tRNA['sequence']
                    )

        # mark ncRNA-regions overlapping with ncRNA-regions
        for ncRNA_region in sequence_ncrna_regions[seq['id']]:
            for ncRNA_region_overlap in sequence_ncrna_regions[seq['id']]:
                if(ncRNA_region['stop'] < ncRNA_region_overlap['start'] or ncRNA_region['start'] > ncRNA_region_overlap['stop']):
                    continue
                if(ncRNA_region['db_xrefs'][0] == ncRNA_region_overlap['db_xrefs'][0]):
                    continue  # same
                else:  # overlap -> select ncRNA based on bitscore, discard the other
                    if(ncRNA_region['score'] < ncRNA_region_overlap['score']):
                        overlap = f"[{max(ncRNA_region['start'], ncRNA_region_overlap['start'])},{min(ncRNA_region['stop'], ncRNA_region_overlap['stop'])}]"
                        ncRNA_region['discarded'] = {
                            'type': bc.DISCARD_TYPE_OVERLAP,
                            'feature_type': bc.FEATURE_NC_RNA_REGION,
                            'description': f"{bc.FEATURE_NC_RNA_REGION} overlap with ({ncRNA_region_overlap['product']}) at {overlap}"
                        }
                        log.info(
                            "overlap: ncRNA-region (%s) [%i, %i] overlapping with ncRNA-region (%s) [%i, %i] at %s on seq=%s, lower bitscore (%f/%f)",
                            ncRNA_region['product'], ncRNA_region['start'], ncRNA_region['stop'], ncRNA_region_overlap['product'], ncRNA_region_overlap['start'], ncRNA_region_overlap['stop'], overlap, ncRNA_region['sequence'], ncRNA_region['score'], ncRNA_region_overlap['score']
                        )

        # mark de novo-predicted CDS overlapping with tRNAs, tmRNAs, rRNAs, CRISPRs and user-provided CDS
        for cds in sequence_cdss[seq['id']]:
            # tmRNA overlaps
            for tmRNA in sequence_tm_rnas[seq['id']]:
                if(cds['stop'] < tmRNA['start'] or cds['start'] > tmRNA['stop']):
                    continue
                else:  # overlap -> remove cds
                    overlap = f"[{max(cds['start'], tmRNA['start'])},{min(cds['stop'], tmRNA['stop'])}]"
                    cds['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_TM_RNA,
                        'description': f"{bc.FEATURE_TM_RNA} overlap with ({tmRNA['product']}) at {overlap}"
                    }
                    log.info(
                        "overlap: CDS (%s/%s) [%i, %i] overlapping tmRNA (%s) [%i, %i], %s, seq=%s",
                        cds.get('gene', '-'), cds.get('product', '-'), cds['start'], cds['stop'], tmRNA['gene'], tmRNA['start'], tmRNA['stop'], overlap, cds['sequence']
                    )
            # tRNA overlaps
            for tRNA in sequence_t_rnas[seq['id']]:
                if(cds['stop'] < tRNA['start'] or cds['start'] > tRNA['stop']):
                    continue
                else:  # overlap -> remove cds
                    overlap = f"[{max(cds['start'], tRNA['start'])},{min(cds['stop'], tRNA['stop'])}]"
                    cds['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_T_RNA,
                        'description': f"{bc.FEATURE_T_RNA} overlap with ({tRNA['product']}) at {overlap}"
                    }
                    log.info(
                        "overlap: CDS (%s/%s) [%i, %i] overlapping tRNA (%s) [%i, %i], %s, seq=%s",
                        cds.get('gene', '-'), cds.get('product', '-'), cds['start'], cds['stop'], tRNA['gene'], tRNA['start'], tRNA['stop'], overlap, cds['sequence']
                    )
            # rRNA overlaps
            for rRNA in sequence_r_rnas[seq['id']]:
                if(cds['stop'] < rRNA['start'] or cds['start'] > rRNA['stop']):
                    continue
                else:  # overlap -> remove cds
                    overlap = f"[{max(cds['start'], rRNA['start'])},{min(cds['stop'], rRNA['stop'])}]"
                    cds['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_R_RNA,
                        'description': f"{bc.FEATURE_R_RNA} overlap with ({rRNA['product']}) at {overlap}"
                    }
                    log.info(
                        "overlap: CDS (%s/%s) [%i, %i] overlapping rRNA (%s) [%i, %i], %s, seq=%s",
                        cds.get('gene', '-'), cds.get('product', '-'), cds['start'], cds['stop'], rRNA['gene'], rRNA['start'], rRNA['stop'], overlap, cds['sequence']
                    )
            # CRISPR overlaps
            for crispr in sequence_crispr_arrays[seq['id']]:
                if(cds['stop'] < crispr['start'] or cds['start'] > crispr['stop']):
                    continue
                else:  # overlap -> remove cds
                    overlap = f"[{max(cds['start'], crispr['start'])},{min(cds['stop'], crispr['stop'])}]"
                    cds['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_CRISPR,
                        'description': f'overlaps {bc.FEATURE_CRISPR} at {overlap}'
                    }
                    log.info(
                        "overlap: CDS (%s/%s) [%i, %i] overlapping CRISPR [%i, %i], %s, seq=%s",
                        cds.get('gene', '-'), cds.get('product', '-'), cds['start'], cds['stop'], crispr['start'], crispr['stop'], overlap, cds['sequence']
                    )
            # user-provided CDS overlaps
            for cds_user_provided in sequence_cdss_user_provided[seq['id']]:
                overlap = 0
                if(not cds_user_provided.get('edge', False)  and  not cds.get('edge', False)):  # both CDS not edge features
                    if(cds['stop'] < cds_user_provided['start'] or cds['start'] > cds_user_provided['stop']):
                        continue
                    else:
                        overlap = min(cds['stop'], cds_user_provided['stop']) - max(cds['start'], cds_user_provided['start']) + 1
                elif(cds_user_provided.get('edge', False)  and  not cds.get('edge', False)):  # only user-provided CDS edge feature
                    if(cds['stop'] > cds_user_provided['start']):  # overlapping de-novo CDS at sequence end
                        overlap = cds['stop'] - max(cds['start'], cds_user_provided['start']) + 1
                    elif(cds['start'] < cds_user_provided['stop']):  # overlapping de-novo CDS at sequence start
                        overlap = min(cds['stop'], cds_user_provided['stop']) - cds['start'] + 1
                    else:
                        continue
                elif(not cds_user_provided.get('edge', False)  and  cds.get('edge', False)):  # only de-novo CDS edge feature
                    if(cds_user_provided['stop'] > cds['start']):  # overlapping user-provided CDS at sequence end
                        overlap = cds_user_provided['stop'] - max(cds['start'], cds_user_provided['start']) + 1
                    elif(cds_user_provided['start'] < cds['stop']):  # overlapping user-provided CDS at sequence start
                        overlap = min(cds['stop'], cds_user_provided['stop']) - cds_user_provided['start'] + 1
                    else:
                        continue
                elif(cds_user_provided.get('edge', False)  and  cds.get('edge', False)):  # both CDS edge features
                    overlap = (seq['length'] - max(cds['start'], cds_user_provided['start']) + 1) + min(cds['stop'], cds_user_provided['stop'])
                if(overlap > bc.CDS_MAX_OVERLAPS):
                    overlap = f"[{max(cds['start'], cds_user_provided['start'])},{min(cds['stop'], cds_user_provided['stop'])}]"
                    cds['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_CDS,
                        'description': f'overlaps user-provided {bc.FEATURE_CDS} at {overlap}'
                    }
                    log.info(
                        "overlap: de-novo CDS (%s/%s) [%i, %i] overlapping user-provided CDS [%i, %i], %s, seq=%s",
                        cds.get('gene', '-'), cds.get('product', '-'), cds['start'], cds['stop'], cds_user_provided['start'], cds_user_provided['stop'], overlap, cds['sequence']
                    )

        # remove sORF overlapping with tRNAs, tmRNAs, rRNAs, CRISPRs, inframe CDSs, shorter inframe sORFs
        for sorf in sequence_sorfs[seq['id']]:
            # tmRNA overlaps
            for tmRNA in sequence_tm_rnas[seq['id']]:
                if(sorf['stop'] < tmRNA['start'] or sorf['start'] > tmRNA['stop']):
                    continue
                else:  # overlap -> remove sorf
                    overlap = f"[{max(sorf['start'], tmRNA['start'])},{min(sorf['stop'], tmRNA['stop'])}]"
                    sorf['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_TM_RNA,
                        'description': f"{bc.FEATURE_TM_RNA} overlap with ({tmRNA['product']}) at {overlap}"
                    }
                    log.info(
                        "overlap: sORF (%s/%s) [%i, %i] overlapping tmRNA (%s) [%i, %i], %s, seq=%s",
                        sorf.get('gene', '-'), sorf.get('product', '-'), sorf['start'], sorf['stop'], tmRNA['gene'], tmRNA['start'], tmRNA['stop'], overlap, sorf['sequence']
                    )
            # tRNA overlaps
            for tRNA in sequence_t_rnas[seq['id']]:
                if(sorf['stop'] < tRNA['start'] or sorf['start'] > tRNA['stop']):
                    continue
                else:  # overlap -> remove sorf
                    overlap = f"[{max(sorf['start'], tRNA['start'])},{min(sorf['stop'], tRNA['stop'])}]"
                    sorf['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_T_RNA,
                        'description': f"{bc.FEATURE_T_RNA} overlap with ({tRNA['product']}) at {overlap}"
                    }
                    log.info(
                        "overlap: sORF (%s/%s) [%i, %i] overlapping tRNA (%s) [%i, %i], %s, seq=%s",
                        sorf.get('gene', '-'), sorf.get('product', '-'), sorf['start'], sorf['stop'], tRNA['gene'], tRNA['start'], tRNA['stop'], overlap, sorf['sequence']
                    )
            # rRNA overlaps
            for rRNA in sequence_r_rnas[seq['id']]:
                if(sorf['stop'] < rRNA['start'] or sorf['start'] > rRNA['stop']):
                    continue
                else:  # overlap -> remove sorf
                    overlap = f"[{max(sorf['start'], rRNA['start'])},{min(sorf['stop'], rRNA['stop'])}]"
                    sorf['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_R_RNA,
                        'description': f"{bc.FEATURE_R_RNA} overlap with ({rRNA['product']}) at {overlap}"
                    }
                    log.info(
                        "overlap: sORF (%s/%s) [%i, %i] overlapping rRNA (%s) [%i, %i], %s, seq=%s",
                        sorf.get('gene', '-'), sorf.get('product', '-'), sorf['start'], sorf['stop'], rRNA['gene'], rRNA['start'], rRNA['stop'], overlap, sorf['sequence']
                    )
            # CRISPR overlaps
            for crispr in sequence_crispr_arrays[seq['id']]:
                if(sorf['stop'] < crispr['start'] or sorf['start'] > crispr['stop']):
                    continue
                else:  # overlap -> remove sorf
                    overlap = f"[{max(sorf['start'], crispr['start'])},{min(sorf['stop'], crispr['stop'])}]"
                    sorf['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_CRISPR,
                        'description': f'overlaps {bc.FEATURE_CRISPR} at {overlap}'
                    }
                    log.info(
                        "overlap: sORF (%s/%s) [%i, %i] overlapping CRISPR [%i, %i], %s, seq=%s",
                        sorf.get('gene', '-'), sorf.get('product', '-'), sorf['start'], sorf['stop'], crispr['start'], crispr['stop'], overlap, sorf['sequence']
                    )
            # user-provided CDS overlaps
            for cds_user_provided in sequence_cdss_user_provided[seq['id']]:
                if(sorf['stop'] < cds_user_provided['start'] or sorf['start'] > cds_user_provided['stop']):
                    continue
                else:  # overlap -> remove sorf
                    overlap = f"[{max(sorf['start'], cds_user_provided['start'])},{min(sorf['stop'], cds_user_provided['stop'])}]"
                    sorf['discarded'] = {
                        'type': bc.DISCARD_TYPE_OVERLAP,
                        'feature_type': bc.FEATURE_CDS,
                        'description': f'overlaps {bc.FEATURE_CDS} at {overlap}'
                    }
                    log.info(
                        "overlap: sORF (%s/%s) [%i, %i] overlapping user-provided CDS [%i, %i], %s, seq=%s",
                        sorf.get('gene', '-'), sorf.get('product', '-'), sorf['start'], sorf['stop'], cds_user_provided['start'], cds_user_provided['stop'], overlap, sorf['sequence']
                    )

            # sORF overlaps
            for overlap_sorf in sequence_sorfs[seq['id']]:
                if(sorf['stop'] < overlap_sorf['start'] or sorf['start'] > overlap_sorf['stop']):
                    continue  # no overlap
                elif(sorf['start'] == overlap_sorf['start'] and sorf['stop'] == overlap_sorf['stop']):
                    continue  # same
                else:  # overlap -> remove sorf
                    score_sorf = calc_cds_annotation_score(sorf)
                    score_overlap_sorf = calc_cds_annotation_score(overlap_sorf)

                    if(score_sorf < score_overlap_sorf):  # lower annotation score
                        overlap = f"[{max(sorf['start'], overlap_sorf['start'])},{min(sorf['stop'], overlap_sorf['stop'])}]"
                        sorf['discarded'] = {
                            'type': bc.DISCARD_TYPE_OVERLAP,
                            'feature_type': bc.FEATURE_SORF,
                            'description': f"overlaps {bc.FEATURE_SORF} ({overlap_sorf.get('gene', '-')}/{overlap_sorf.get('product', '-')}) at {overlap} with lower score ({score_sorf}/{score_overlap_sorf})"
                        }
                        log.info(
                            "overlap: sORF (%s/%s) [%i, %i] overlapping sORF (%s/%s) [%i, %i], %s, seq=%s, lower annotation score (%i/%i)",
                            sorf.get('gene', '-'), sorf.get('product', '-'), sorf['start'], sorf['stop'], overlap_sorf.get('gene', '-'), overlap_sorf.get('product', '-'), overlap_sorf['start'], overlap_sorf['stop'], overlap, sorf['sequence'], score_sorf, score_overlap_sorf
                        )
                    elif(score_sorf == score_overlap_sorf and len(sorf['aa']) < len(overlap_sorf['aa'])):  # equal annotation score but shorter sequence -> potential fragment or too short ORF prediction
                        overlap = f"[{max(sorf['start'], overlap_sorf['start'])},{min(sorf['stop'], overlap_sorf['stop'])}]"
                        sorf['discarded'] = {
                            'type': bc.DISCARD_TYPE_OVERLAP,
                            'feature_type': bc.FEATURE_SORF,
                            'description': f"overlaps {bc.FEATURE_SORF} ({overlap_sorf.get('gene', '-')}/{overlap_sorf.get('product', '-')}) at {overlap} with equal score ({score_sorf}) but lower length ({len(sorf['aa'])}/{len(overlap_sorf['aa'])})"
                        }
                        log.info(
                            "overlap: sORF (%s/%s) [%i, %i] overlapping sORF (%s/%s) [%i, %i], %s, seq=%s, equal annotation score (%i), lower length (%i/%i)",
                            sorf.get('gene', '-'), sorf.get('product', '-'), sorf['start'], sorf['stop'], overlap_sorf.get('gene', '-'), overlap_sorf.get('product', '-'), overlap_sorf['start'], overlap_sorf['stop'], overlap, sorf['sequence'], score_sorf, len(sorf['aa']), len(overlap_sorf['aa'])
                        )


def calc_cds_annotation_score(cds: dict) -> int:
    """Calc an annotation score rewarding each identification & annotation"""
    score = 0

    if('ups' in cds):
        score += 1

    ips = cds.get('ips', None)
    if(ips):
        score += 1
        score += calc_annotation_score(ips)

    psc = cds.get('psc', None)
    if(psc):
        score += 1
        score += calc_annotation_score(psc)
    log.debug(
        'cds score: seq=%s, start=%i, stop=%i, gene=%s, product=%s, score=%i',
        cds['sequence'], cds['start'], cds['stop'], cds.get('gene', '-'), cds.get('product', '-'), score
    )
    return score


def calc_annotation_score(orf:dict) -> int:
    score = 0
    if(orf.get('gene', None)):
        score += 1
    if(orf.get('product', None)):
        score += 1
    return score


def extract_protein_gene_symbol(product: str) -> str:
    gene_symbols = []
    for part in product.split(' '):  # try to extract valid gene symbols
        m = RE_GENE_SYMBOL.fullmatch(part)
        if(m):
            symbol = m[0]
            log.info('fix gene: extract symbol from protein name. symbol=%s', symbol)
            gene_symbols.append(symbol)
        else:
            m = RE_PROTEIN_SYMBOL.fullmatch(part)  # extract protein names
            if(m):
                symbol = m[0]
                symbol = symbol[0].lower() + symbol[1:]
                log.info('fix gene: extract symbol from protein name. symbol=%s', symbol)
                gene_symbols.append(symbol)
    if(len(gene_symbols) == 0):  # None found
        return None
    elif(len(gene_symbols) == 1):  # found 1
        return gene_symbols[0]
    else:  # found more than one, take the 2nd as the 1st often describes a broader gene family like "xyz family trancsriptional regulator ..."
        return gene_symbols[1]


def revise_cds_gene_symbols(raw_genes: Sequence[str]):
    revised_genes = set()
    for gene in raw_genes:
        old_gene = gene
        if(RE_GENE_SUSPECT_CHARS.search(gene)):  # check for suspect characters -> remove gene symbol
            log.info('fix gene: remove gene symbol containing suspect chars. old=%s', old_gene)
            continue

        old_gene = gene
        gene = gene.replace('gene', '')
        if(gene != old_gene):  # remove gene literal
            log.info('fix gene: remove gene literal. new=%s, old=%s', gene, old_gene)

        old_gene = gene
        if(gene[-1] == '-'):  # remove orphan hyphen
            gene = gene[:-1]
            log.info('fix gene: remove orphan hypen. new=%s, old=%s', gene, old_gene)
        
        old_gene = gene
        gene = RE_MULTIWHITESPACE.sub(' ', gene).strip()  # revise whitespaces
        if(gene != old_gene):
            log.info('fix gene: revise whitespaces. new=%s, old=%s', gene, old_gene)

        old_gene = gene
        if(RE_GENE_CAPITALIZED.fullmatch(gene)):
            gene = gene[0].lower() + gene[1:]
            log.info('fix gene: lowercase first char. new=%s, old=%s', gene, old_gene)

        if(len(gene) >= 3):
            if(len(gene) <= 12):
                revised_genes.add(gene)
            else:
                old_gene = gene
                gene = extract_protein_gene_symbol(gene)
                if(gene):
                    revised_genes.add(gene)
    return list(revised_genes)


def revise_cds_product(product: str):
    """Revise product name for INSDC compliant submissions"""
    
    old_product = product
    product = RE_PROTEIN_WEIGHT.sub(' ', product)  # remove protein weight in (k)Da
    if(product != old_product):
        log.info('fix product: remove protein weight in (k)Da. new=%s, old=%s', product, old_product)

    old_product = product
    product = re.sub(RE_PROTEIN_PERIOD_SEPARATOR, r'\1-\2', product)  # replace separator periods
    if(product != old_product):
        log.info('fix product: replace separator periods. new=%s, old=%s', product, old_product)
    
    old_product = product
    if(product[0] in RE_PROTEIN_SUSPECT_CHARS_BEGINNING):  # remove suspect first character
        product = product[1:]
        log.info('fix product: replace invalid first character. new=%s, old=%s', product, old_product)

    old_product = product
    product = RE_PROTEIN_SUSPECT_CHARS_DISCARD.sub('', product)  # remove suspect characters
    if(product != old_product):
        log.info('fix product: replace invalid characters. new=%s, old=%s', product, old_product)

    old_product = product
    product = RE_PROTEIN_SUSPECT_CHARS_REPLACE.sub(' ', product)  # replace suspect characters by single whitespace
    if(product != old_product):
        log.info('fix product: replace invalid characters. new=%s, old=%s', product, old_product)

    old_product = product
    product = RE_PROTEIN_WRONG_PRIMES.sub('\u0027', product)  # replace wrong prime characters with single quote (U+0027) (') according to https://www.ncbi.nlm.nih.gov/genome/doc/internatprot_nomenguide/
    if(product != old_product):
        log.info('fix product: replace wrong prime characters. new=%s, old=%s', product, old_product)

    old_product = product
    product = product.replace('FOG:', '')  # remove FOG ids
    if(product != old_product):
        log.info('fix product: replace FOG ids. new=%s, old=%s', product, old_product)

    old_product = product
    product = RE_PROTEIN_REMNANT.sub('', product)  # remove 'Remnant of's
    if(product != old_product):
        log.info('fix product: replace remnant ofs. new=%s, old=%s', product, old_product)

    old_product = product
    dufs = []  # replace DUF-containing products
    for m in RE_DOMAIN_OF_UNKNOWN_FUNCTION.finditer(product):
        dufs.append(m.group(1).upper())
    if(len(dufs) >= 1):
        product = f"{' '.join(dufs)} domain{'s' if len(dufs) > 1 else ''}-containing protein"
        if(product != old_product):
            log.info('fix product: revise DUF. new=%s, old=%s', product, old_product)
    
    old_product = product
    if('conserved' in product.lower()):  # replace conserved UPF proteins
        upfs = []
        for m in RE_UNCHARACTERIZED_PROTEIN_FAMILY.finditer(product):
            upfs.append(m.group(1).upper())
        if(len(upfs) >= 1):
            product = f"{' '.join(upfs)} protein"
            if(product != old_product):
                log.info('fix product: revise UPF. new=%s, old=%s', product, old_product)

    old_product = product
    product = RE_PROTEIN_HOMOLOG.sub('-like protein', product)  # replace Homologs
    if(product != old_product):
        if(product.count('protein') == 2):
            product = product.replace('protein', '', 1)  # remove former protein term if existing
        log.info('fix product: replace Homolog. new=%s, old=%s', product, old_product)

    old_product = product
    product = RE_MULTIWHITESPACE.sub(' ', product).strip()  # revise whitespaces
    if(product != old_product):
        log.info('fix product: revise whitespaces. new=%s, old=%s', product, old_product)

    old_product = product
    product = RE_PROTEIN_PUTATIVE.sub('putative', product)  # replace putative synonyms)
    if(product != old_product):
        log.info('fix product: replace putative synonyms. new=%s, old=%s', product, old_product)

    old_product = product
    if(RE_PROTEIN_DOMAIN_CONTAINING.search(product)):  # replace domain name underscores in domain names
        product = product.replace('_', '-')
        if(product != old_product):
            log.info('fix product: replace domain name underscores. new=%s, old=%s', product, old_product)
    
    old_product = product
    if(RE_PROTEIN_TMRNA.fullmatch(product)):
        product = ''
        log.info('fix product: discard pure tmRNA product descriptions. new=%s, old=%s', product, old_product)

    old_product = product
    if(
        RE_PROTEIN_CONTIG.search(product) or  # protein containing 'sequence'
        RE_PROTEIN_NODE.search(product) or  # potential contig name (SPAdes)
        RE_PROTEIN_POTENTIAL_CONTIG_NAME.search(product) or  # potential contig name (SPAdes)
        RE_PROTEIN_NO_LETTERS.fullmatch(product)  # no letters -> set to Hypothetical
        ):  # remove suspect products and mark as hypothetical
        product = None
        log.info('remove product: mark proteins with suspect products as hypothetical. old=%s', old_product)

    return product


def mark_as_hypothetical(feature: dict):
    log.info(
        'marked as hypothetical: seq=%s, start=%i, stop=%i, strand=%s',
        feature['sequence'], feature['start'], feature['stop'], feature['strand']
    )
    feature['hypothetical'] = True
    feature['gene'] = None
    feature['genes'] = []
    feature['product'] = bc.HYPOTHETICAL_PROTEIN


def get_adjacent_genes(feature: dict, features: Sequence[dict], neighbors=3):
    for idx, feat in enumerate(features):
        if feat['locus'] == feature['locus']:
            upstream_genes = []
            if(idx >= 1):
                start = idx - neighbors
                if(start < 0 ):
                    start = 0
                upstream_genes = features[start:idx]
            downstream_genes = []
            if(idx + 1 < len(features)):
                end = idx + 1 + neighbors
                if(end > len(features)):
                    end = len(features)
                downstream_genes = features[idx+1:end]
            upstream_genes.extend(downstream_genes)
            for gene in upstream_genes:
                log.debug(
                    'extracted neighbor genes: seq=%s, start=%i, stop=%i, gene=%s, product=%s',
                    gene['sequence'], gene['start'], gene['stop'], gene.get('gene', '-'), gene.get('product', '-')
                )
            return upstream_genes
    return []


def select_gene_symbols(features: Sequence[dict]):
    improved_genes = []
    for feat in [f for f in features if len(f.get('genes', [])) > 1]:  # all CDS/sORF with multiple gene symbols
        old_gene_symbol = feat['gene']
        gene_symbol_prefixes = set([symbol[:3] for symbol in feat['genes'] if len(symbol) > 3])
        if(len(gene_symbol_prefixes) == 1 ):  # multiple gene symbols of the same prefix: 
            product_parts = feat.get('product', '').split()
            for gene_symbol in feat['genes']:
                protein_symbol = gene_symbol[0].upper() + gene_symbol[1:]
                if(protein_symbol == product_parts[-1]):  # gene symbol is last part of product description which often is a specific gene/protein name
                    if(gene_symbol != old_gene_symbol):
                        feat['gene'] = gene_symbol
                        log.info(
                            'gene product symbol selection: seq=%s, start=%i, stop=%i, new-gene=%s, old-gene=%s, genes=%s, product=%s',
                            feat['sequence'], feat['start'], feat['stop'], gene_symbol, old_gene_symbol, ','.join(feat['genes']), feat.get('product', '-')
                        )
                        improved_genes.append(feat)
        else:  # multiple gene symbols of varying prefixes are available, e.g. acrS, envR
            log.debug(
                'select gene symbol: seq=%s, start=%i, stop=%i, gene=%s, genes=%s, product=%s',
                feat['sequence'], feat['start'], feat['stop'], feat.get('gene', '-'), ','.join(feat['genes']), feat.get('product', '-')
            )
            adjacent_genes = get_adjacent_genes(feat, features, neighbors=3)
            adjacent_gene_symbol_lists = [gene.get('genes', []) for gene in adjacent_genes]
            adjacent_gene_symbols = [item for sublist in adjacent_gene_symbol_lists for item in sublist]  # flatten lists
            adjacent_gene_symbol_prefixes = [gene_symbol[:3] for gene_symbol in adjacent_gene_symbols if len(gene_symbol) > 3]  # extract gene symbol prefixes, e.g. tra for traI, traX, traM
            adjacent_gene_symbol_prefix_counts = {}
            for gene_symbol_prefix in adjacent_gene_symbol_prefixes:
                if gene_symbol_prefix in adjacent_gene_symbol_prefix_counts:
                    adjacent_gene_symbol_prefix_counts[gene_symbol_prefix] += 1
                else:
                    adjacent_gene_symbol_prefix_counts[gene_symbol_prefix] = 1
            log.debug('neighbor gene symbol prefix counts: %s', adjacent_gene_symbol_prefix_counts)
            count = 0
            selected_gene_symbol = old_gene_symbol
            for gene_symbol in feat['genes']:
                gene_symbol_prefix = gene_symbol[:3]
                gene_symbol_count = adjacent_gene_symbol_prefix_counts.get(gene_symbol_prefix, 0)
                if gene_symbol_count > count:  # select symbol if its prefix is dominant in the gene neighborhood (neihboorhood of 3 genes up-/downstream as operon proxy)
                    selected_gene_symbol = gene_symbol
                    count = gene_symbol_count
            if(selected_gene_symbol != old_gene_symbol):
                feat['gene'] = selected_gene_symbol
                log.info(
                    'gene neighborhood symbol selection: seq=%s, start=%i, stop=%i, new-gene=%s, old-gene=%s, genes=%s, product=%s',
                    feat['sequence'], feat['start'], feat['stop'], selected_gene_symbol, old_gene_symbol, ','.join(feat['genes']), feat.get('product', '-')
                )
                improved_genes.append(feat)
    return improved_genes


import argparse
import collections
import csv
import hashlib
import logging
import os
import platform as pf
import re
import shutil
import sys
import subprocess as sp
import sqlite3
import threading

from argparse import Namespace
from datetime import datetime
from typing import Dict, Sequence, Tuple
from pathlib import Path

from Bio.Seq import Seq

import bakta
import bakta.constants as bc
import bakta.config as cfg


log = logging.getLogger('UTILS')


def print_version(self):
    return f'v{self.major}.{self.minor}.{self.patch}'


Version = collections.namedtuple('Version', ['major', 'minor', 'patch'], defaults=[0, 0])  # named tuple for version checking, defaults are zero for missing minor/patch
Version.__str__ = print_version


VERSION_MIN_DIGIT = -1
VERSION_MAX_DIGIT = 1000000000000
VERSION_REGEX = re.compile(r'(\d+)\.(\d+)(?:[\.-](\d+))?')  # regex to search for version number in tool output. Takes missing patch version into consideration.

# List of dependencies: tuples for: min version, max version, tool name & command line parameter, dependency check exclusion options
DEPENDENCY_TRNASCAN = (Version(2, 0, 12), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, 'tRNAscan-SE', ('tRNAscan-SE', '-h'), ['--skip-trna'])
DEPENDENCY_ARAGORN = (Version(1, 2, 41), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, 'Aragorn', ('aragorn', '-h'), ['skip-tmrna'])
DEPENDENCY_CMSCAN = (Version(1, 1, 5), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, 'CMscan', ('cmscan', '-h'), ['--skip-rrna', '--skip-ncrna', '--skip-ncrna-region'])
DEPENDENCY_PILERCR = (Version(1, 6), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, 'PilerCR', ('pilercr', '-options'), ['--skip-crispr'])
DEPENDENCY_PYRODIGAL = (Version(3, 7, 0), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, 'Pyrodigal', (sys.executable, '-c', 'import pyrodigal; print(pyrodigal.__version__)'), ['--skip-cds'])
DEPENDENCY_PYHMMER = (Version(0, 12, 0), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, 'Pyhmmer', (sys.executable, '-c', 'import pyhmmer; print(pyhmmer.__version__)'), ['--skip-cds', '--skip-sorf'])
DEPENDENCY_DIAMOND = (Version(2, 1, 21), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, 'Diamond', ('diamond', 'help'), ['--skip-cds', '--skip-sorf'])
DEPENDENCY_BLASTN = (Version(2, 17, 0), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, 'Blastn', ('blastn', '-version'), ['--skip-ori'])
DEPENDENCY_AMRFINDERPLUS = (Version(4, 2, 7), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, 'AMRFinderPlus', ('amrfinder', '--version'), ['--skip-cds'])
DEPENDENCY_PYCIRCLIZE = (Version(1, 7, 0), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, 'pyCirclize', (sys.executable, '-c', 'import pycirclize; print(pycirclize.__version__)'), ['--skip-plot'])


_local = threading.local()  # # init thread-local storage for DB connections


def init_parser(sub_command: str=''):
    parser = argparse.ArgumentParser(
        prog=f'bakta{sub_command}',
        description='Rapid & standardized annotation of bacterial genomes, MAGs & plasmids',
        epilog=f'Version: {cfg.version}\nDOI: {bc.BAKTA_DOI}\nURL: github.com/oschwengers/bakta\n\nCitation:\n{bc.BAKTA_CITATION}',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False
    )
    return parser


def parse_arguments():
    parser = init_parser()
    parser.add_argument('genome', metavar='<genome>', help='Genome sequences in (zipped) fasta format')

    arg_group_io = parser.add_argument_group('Input / Output')
    arg_group_io.add_argument('--db', '-d', action='store', default=None, help='Database path (default = <bakta_path>/db). Can also be provided as BAKTA_DB environment variable.')
    arg_group_io.add_argument('--min-contig-length', '-m', action='store', type=int, default=1, dest='min_contig_length', help='Minimum contig/sequence size (default = 1; 200 in compliant mode)')
    arg_group_io.add_argument('--prefix', '-p', action='store', default=None, help='Prefix for output files')
    arg_group_io.add_argument('--output', '-o', action='store', default=os.getcwd(), help='Output directory (default = current working directory)')
    arg_group_io.add_argument('--force', '-f', action='store_true', help='Force overwriting existing output folder (except for current working directory)')

    arg_group_organism = parser.add_argument_group('Organism')
    arg_group_organism.add_argument('--genus', action='store', default=None, help='Genus name')
    arg_group_organism.add_argument('--species', action='store', default=None, help='Species name')
    arg_group_organism.add_argument('--strain', action='store', default=None, help='Strain name')
    arg_group_organism.add_argument('--plasmid', action='store', default=None, help='Plasmid name')

    arg_group_annotation = parser.add_argument_group('Annotation')
    arg_group_annotation.add_argument('--complete', action='store_true', help='All sequences are complete replicons (chromosome/plasmid[s])')
    arg_group_annotation.add_argument('--prodigal-tf', action='store', default=None, dest='prodigal_tf', help='Path to existing Prodigal training file to use for CDS prediction')
    arg_group_annotation.add_argument('--translation-table', action='store', type=int, default=11, choices=[11, 4, 25], dest='translation_table', help='Translation table: 11/4/25 (default = 11)')
    arg_group_annotation.add_argument('--gram', action='store', default=bc.GRAM_UNKNOWN, choices=[bc.GRAM_POSITIVE, bc.GRAM_NEGATIVE, bc.GRAM_UNKNOWN], help=f'Gram type for signal peptide predictions: {bc.GRAM_POSITIVE}/{bc.GRAM_NEGATIVE}/{bc.GRAM_UNKNOWN} (default = {bc.GRAM_UNKNOWN})')
    arg_group_annotation.add_argument('--locus', action='store', default=None, help="Locus prefix (default = 'contig')")
    arg_group_annotation.add_argument('--locus-tag', action='store', default=None, dest='locus_tag', help='Locus tag prefix (default = autogenerated)')
    arg_group_annotation.add_argument('--locus-tag-increment', action='store', type=int, default=1, choices=[1, 5, 10], dest='locus_tag_increment', help='Locus tag increment: 1/5/10 (default = 1)')
    arg_group_annotation.add_argument('--keep-contig-headers', action='store_true', dest='keep_contig_headers', help='Keep original contig/sequence headers')
    arg_group_annotation.add_argument('--compliant', action='store_true', help='Force Genbank/ENA/DDJB compliance')
    arg_group_annotation.add_argument('--replicons', '-r', action='store', default=None, dest='replicons', help='Replicon information table (tsv/csv)')
    arg_group_annotation.add_argument('--regions', action='store', default=None, help='Path to pre-annotated regions in GFF3 or Genbank format (regions only, no functional annotations).')
    arg_group_annotation.add_argument('--proteins', action='store', default=None, dest='proteins', help='Fasta file of trusted protein sequences for CDS annotation')
    arg_group_annotation.add_argument('--hmms', action='store', default=None, dest='hmms', help='HMM file of trusted hidden markov models in HMMER format for CDS annotation')
    arg_group_annotation.add_argument('--meta', action='store_true', help='Run in metagenome mode. This only affects CDS prediction.')
    arg_group_annotation.add_argument('--partial', action='store_true', help='Predict partial (truncated) genes spanning linear sequence ends')

    arg_group_workflow = parser.add_argument_group('Workflow')
    arg_group_workflow.add_argument('--skip-trna', action='store_true', dest='skip_trna', help='Skip tRNA detection & annotation')
    arg_group_workflow.add_argument('--skip-tmrna', action='store_true', dest='skip_tmrna', help='Skip tmRNA detection & annotation')
    arg_group_workflow.add_argument('--skip-rrna', action='store_true', dest='skip_rrna', help='Skip rRNA detection & annotation')
    arg_group_workflow.add_argument('--skip-ncrna', action='store_true', dest='skip_ncrna', help='Skip ncRNA detection & annotation')
    arg_group_workflow.add_argument('--skip-ncrna-region', action='store_true', dest='skip_ncrna_region', help='Skip ncRNA region detection & annotation')
    arg_group_workflow.add_argument('--skip-crispr', action='store_true', dest='skip_crispr', help='Skip CRISPR array detection & annotation')
    arg_group_workflow.add_argument('--skip-cds', action='store_true', dest='skip_cds', help='Skip CDS detection & annotation')
    arg_group_workflow.add_argument('--skip-pseudo', action='store_true', dest='skip_pseudo', help='Skip pseudogene detection & annotation')
    arg_group_workflow.add_argument('--skip-sorf', action='store_true', dest='skip_sorf', help='Skip sORF detection & annotation')
    arg_group_workflow.add_argument('--skip-gap', action='store_true', dest='skip_gap', help='Skip gap detection & annotation')
    arg_group_workflow.add_argument('--skip-ori', action='store_true', dest='skip_ori', help='Skip oriC/oriT detection & annotation')
    arg_group_workflow.add_argument('--skip-filter', action='store_true', dest='skip_filter', help='Skip feature overlap filters')
    arg_group_workflow.add_argument('--skip-plot', action='store_true', dest='skip_plot', help='Skip generation of circular genome plots')

    arg_group_general = parser.add_argument_group('General')
    arg_group_general.add_argument('--help', '-h', action='help', help='Show this help message and exit')
    arg_group_general.add_argument('--verbose', '-v', action='store_true', help='Print verbose information')
    arg_group_general.add_argument('--debug', action='store_true', help='Run Bakta in debug mode. Temp data will not be removed.')
    arg_group_general.add_argument('--threads', '-t', action='store', type=int, default=0, help='Number of threads to use (default = number of available CPUs)')
    arg_group_general.add_argument('--tmp-dir', action='store', default=None, dest='tmp_dir', help='Location for temporary files (default = system dependent auto detection)')
    arg_group_general.add_argument('--version', action='version', version=f'%(prog)s {cfg.version}')
    return parser.parse_args()


def setup_logger(output_path: Path, prefix: str, args: Namespace):
    logging.basicConfig(
        filename=str(output_path.joinpath(f'{prefix}.log')),
        filemode='w',
        format='%(asctime)s.%(msecs)03d - %(levelname)s - %(name)s - %(message)s',
        datefmt='%H:%M:%S',
        level=logging.DEBUG if args.debug else logging.INFO
    )
    log.info('version=%s', cfg.version)
    log.info('developer: Oliver Schwengers, github.com/oschwengers')
    log.info('command: %s', ' '.join(sys.argv))
    log.info('local time: %s', datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
    log.info('machine: type=%s, cores=%s', pf.processor(), os.cpu_count())
    log.info('system: type=%s, release=%s', pf.system(), pf.release())
    log.info('python: version=%s, implementation=%s', pf.python_version(), pf.python_implementation())


def cleanup(log, tmp_path: Path):
    shutil.rmtree(str(tmp_path))  # remove tmp dir
    log.info('removed tmp dir: %s', tmp_path)


def read_tool_output(dependency):
    """Method for reading tool version with regex. Input: regex expression, tool command. Retursn: version number."""
    version_regex = dependency[2]
    tool = dependency[3]
    command = dependency[4]
    skip_options = dependency[5]
    try:
        tool_output = str(sp.check_output(command, stderr=sp.STDOUT))  # stderr must be added in case the tool output is not piped into stdout
        log.debug('tool output: tool=%s, output=%s', tool, tool_output)
    except FileNotFoundError:
        log.error('dependency not found! tool=%s', tool)
        sys.exit(f"ERROR: {tool} not found or not executable! Please make sure {tool} is installed and executable or skip requiring workflow steps via via '{' '.join(skip_options)}'.")
    except sp.CalledProcessError:
        log.error('dependency check failed! tool=%s', tool)
        sys.exit(f"ERROR: {tool} could not be executed! Please make sure {tool} is installed and executable or skip requiring workflow steps via via '{' '.join(skip_options)}'.")
    version_match = version_regex.search(tool_output)
    try:
        if version_match is None:
            log.error('no dependency version detected! No regex hit in dependency output: regex=%s, command=%s', version_regex, command)
            sys.exit(f'ERROR: Could not detect/read {tool} version! No regex hit in dependency output.')
        major = version_match.group(1)
        minor = version_match.group(2)
        patch = version_match.group(3)
        if major is None:
            log.error('no dependency version detected! Major hit is None in dependency output: regex=%s, command=%s', version_regex, command)
            sys.exit(f'ERROR: Could not detect/read {tool} version! Major hit is None in dependency output.')
        elif minor is None:
            version_output = Version(int(major))
        elif patch is None:
            version_output = Version(int(major), int(minor))
        else:
            version_output = Version(int(major), int(minor), int(patch))
        return version_output
    except:
        log.error('no dependency version detected! Error while reading dependency output: regex=%s, command=%s', version_regex, command)
        sys.exit(f'ERROR: Could not detect/read {tool} version! Error while reading dependency output.')


def check_version(tool, min: int, max:int ) -> bool:
    """Method for checking tool versions with required version. Input: tool version, minimum and maximum version. Returns: boolean value for positive or negative check."""
    if tool.major < min.major or tool.major > max.major:
        return False
    else:
        if tool.major == min.major or tool.major == max.major:
            if tool.minor < min.minor and tool.major == min.major:
                return False
            elif tool.minor > max.minor and tool.major == max.major:
                return False
            else:
                if tool.minor == min.minor or tool.minor == max.minor:
                    if tool.patch < min.patch and tool.minor == min.minor:
                        return False
                    elif tool.patch > max.patch and tool.minor == max.minor and tool.major == max.major:
                        return False
                    else:
                        return True
                else:
                    return True
        else:
            return True


def test_dependency(dependency):
    """Test the proper installation of the required 3rd party executable."""
    version = read_tool_output(dependency)
    check_result = check_version(version, dependency[0], dependency[1])
    tool_name = dependency[3]
    if (not check_result):
        log.error('wrong dependency version for %s: installed=%s, minimum=%s', tool_name, version, dependency[0])
        sys.exit(f'ERROR: Wrong {tool_name} version installed. Please, either install {tool_name} version {dependency[0]} or use {dependency[4]}!')
    else:
        tool_path = shutil.which(dependency[4][0])
        log.info('dependency: tool=%s, version=%s, path=%s', tool_name, version, tool_path)


def test_dependencies():
    """Test the proper installation of all required 3rd party executables."""
    if(cfg.skip_trna is not None and cfg.skip_trna is False):
        test_dependency(DEPENDENCY_TRNASCAN)

    if(cfg.skip_tmrna is not None and cfg.skip_tmrna is False):
        test_dependency(DEPENDENCY_ARAGORN)

    if((cfg.skip_rrna is not None and cfg.skip_rrna is False) or (cfg.skip_ncrna is not None and cfg.skip_ncrna is False) or (cfg.skip_ncrna_region is not None and cfg.skip_ncrna_region is False)):
        test_dependency(DEPENDENCY_CMSCAN)

    if(cfg.skip_crispr is not None and cfg.skip_crispr is False):
        test_dependency(DEPENDENCY_PILERCR)

    if(cfg.skip_cds is not None and cfg.skip_cds is False):
        test_dependency(DEPENDENCY_PYRODIGAL)
        test_dependency(DEPENDENCY_AMRFINDERPLUS)

        # test if AMRFinderPlus db is installed
        amrfinderplus_db_path = cfg.db_path.joinpath('amrfinderplus-db')
        amrfinderplus_db_latest_path = amrfinderplus_db_path.joinpath('latest')
        process = sp.run(
            [
                'amrfinder',
                '--debug',
                '--database', str(amrfinderplus_db_latest_path)
            ], capture_output=True)
        if('No valid AMRFinder database found' in process.stderr.decode()):
            log.error('AMRFinderPlus database not installed')
            sys.exit(f"ERROR: AMRFinderPlus database not installed! Please, install AMRFinderPlus's internal database by executing: 'amrfinder_update --database {amrfinderplus_db_path}'. This must be done only once.")

    if((cfg.skip_cds is not None and cfg.skip_cds is False) or (cfg.skip_sorf is not None and cfg.skip_sorf is False)):
        test_dependency(DEPENDENCY_PYHMMER)
        test_dependency(DEPENDENCY_DIAMOND)

    if(cfg.skip_ori is not None and cfg.skip_ori is False):
        test_dependency(DEPENDENCY_BLASTN)
    
    if(cfg.skip_plot is not None and cfg.skip_plot is False):
        test_dependency(DEPENDENCY_PYCIRCLIZE)


def create_locus_tag_prefix(sequences: Sequence[dict], length: int=6) -> str:
    """Create either genus/species or sequence MD5 hex based locus tag prefix."""
    hash = hashlib.md5()
    for seq in sequences:
        hash.update(str.encode(seq['nt']))
    hexdigest = hash.hexdigest().upper()
    locus_prefix_chars = []
    i = 0
    while i < length:
        c = hexdigest[i]
        if(c >= '0' and c <= '9'):
            c = chr(ord('F') + int(c) + 1)
        locus_prefix_chars.append(c)
        i += 1
    locus_prefix = ''.join(locus_prefix_chars)
    log.info('generated sequence tag prefix: prefix=%s, length=%i, MD5=%s', locus_prefix, length, hexdigest)
    return locus_prefix


def calc_aa_hash(seq: str) -> Tuple[bytes, str]:
    aa_hash = hashlib.md5(seq.encode('utf-8'))
    return (aa_hash.digest(), aa_hash.hexdigest())


def has_annotation(feature: dict, attribute: str) -> bool:
    value = feature.get(attribute, None)
    if(value is not None and value != ''):
        return True
    else:
        return False


def calc_genome_stats(data: dict):
    genome_size = data['stats']['size']
    log.info('genome-size=%i', genome_size)

    # N50
    gc_sum = 0
    n_sum = 0
    for seq in data['sequences']:
        nt = seq['nt']
        gc_sum += nt.count('G') + nt.count('C')
        n_sum += nt.count('N')
    gc_ratio = gc_sum / (genome_size - n_sum)
    data['stats']['gc'] = gc_ratio
    log.info('GC=%0.3f', gc_ratio)

    n_ratio = n_sum / genome_size
    data['stats']['n_ratio'] = n_ratio
    log.info('N=%0.3f', n_ratio)

    sequence_length_sum = 0
    for seq in sorted(data['sequences'], key=lambda x: x['length'], reverse=True):
        nt_length = len(seq['nt'])
        sequence_length_sum += nt_length
        if(sequence_length_sum >= genome_size * 0.5):
            if 'n50' not in data['stats']:
                data['stats']['n50'] = nt_length
                log.info('N50=%i', nt_length)
        if(sequence_length_sum >= genome_size * 0.9):
            if 'n90' not in data['stats']:
                data['stats']['n90'] = nt_length
                log.info('N90=%i', nt_length)
    

    sequence_by_id = {seq['id']: seq for seq in data['sequences']}
    coding_nts = 0
    for feat in data['features']:
        if(feat.get('edge', False)):
            sequence_length = sequence_by_id[feat['sequence']]['length']
            coding_nts += feat['stop'] + (sequence_length - feat['start'] + 1)  # feature coding nucleotides
        else:
            coding_nts += feat['stop'] - feat['start'] + 1  # feature coding nucleotides
    coding_ratio = coding_nts / (genome_size - n_sum)
    data['stats']['coding_ratio'] = coding_ratio
    log.info('coding-ratio=%0.3f', coding_ratio)


def parse_replicon_table(replicon_table_path: Path) -> Dict[str, dict]:
    replicons = {}
    try:
        with replicon_table_path.open() as fh:
            dialect = csv.Sniffer().sniff(fh.read(1024), delimiters=",\t")
            fh.seek(0)
            reader = csv.reader(fh, dialect)
            for row in reader:
                (original_locus_id, new_locus_id, replicon_type, topology, name) = row
                original_locus_id = original_locus_id.strip()
                if(' ' in original_locus_id):
                    original_locus_id = original_locus_id.split(' ')[0]  # remove description
                new_locus_id = new_locus_id.strip()
                if(' ' in new_locus_id):
                    new_locus_id = new_locus_id.split(' ')[0]  # remove description

                # TODO: add locus id checks
                if(new_locus_id == '' or new_locus_id == '-'):
                    new_locus_id = None
                replicon_type = replicon_type.lower()
                if(replicon_type == 'c' or 'chrom' in replicon_type):
                    replicon_type = bc.REPLICON_CHROMOSOME
                elif(replicon_type == 'p' or 'plasmid' in replicon_type):
                    replicon_type = bc.REPLICON_PLASMID
                else:
                    replicon_type = bc.REPLICON_CONTIG
                topology = topology.lower()
                if(topology == 'c' or 'circ' in topology):
                    topology = bc.TOPOLOGY_CIRCULAR
                else:
                    topology = bc.TOPOLOGY_LINEAR
                if(replicon_type == bc.REPLICON_CONTIG):
                    topology == bc.TOPOLOGY_LINEAR
                if(name == '' or name == '-'):
                    name = None
                replicon = {
                    'original_locus_id': original_locus_id,
                    'new_locus_id': new_locus_id,
                    'replicon_type': replicon_type,
                    'topology': topology,
                    'name': name
                }
                log.info(
                    'replicon info: orig-id=%s, new-id=%s, type=%s, topology=%s, name=%s',
                    replicon['original_locus_id'], replicon['new_locus_id'], replicon['replicon_type'], replicon['topology'], replicon['name']
                )
                replicons[original_locus_id] = replicon
    except:
        log.error('wrong replicon table format!')
        sys.exit('ERROR: wrong replicon table file format!')
    return replicons


def qc_sequences(sequences: Sequence[dict], replicons: Dict[str, dict]) -> Tuple[Sequence[dict], bool]:
    valid_sequences = []
    sequence_counter = 1
    sequence_prefix = cfg.locus if cfg.locus else 'contig'
    complete_genome = True
    plasmid_number = 1
    sequence_ids = set()
    for seq in sequences:
        if(seq['length'] >= cfg.min_sequence_length):
            sequence_id_generated = f'{sequence_prefix}_{sequence_counter}'
            seq['simple_id'] = sequence_id_generated
            sequence_counter += 1

            sequence_description = seq['description'].lower()
            if(cfg.complete):
                seq['complete'] = True
                seq['topology'] = bc.TOPOLOGY_CIRCULAR
            elif('circular=true' in sequence_description):  # detection of Unicycler circularized sequences
                seq['complete'] = True
                seq['topology'] = bc.TOPOLOGY_CIRCULAR
                log.debug('qc: detected Unicycler circular topology via description: id=%s, description=%s', seq['id'], seq['description'])
            elif('complete' in sequence_description and 'complete=false' not in sequence_description):  # detection of public/described sequences
                seq['complete'] = True
                seq['topology'] = bc.TOPOLOGY_CIRCULAR
                log.debug('qc: detected complete replicon via description: id=%s, description=%s', seq['id'], seq['description'])
            
            if('chromosome' in sequence_description):
                seq['type'] = bc.REPLICON_CHROMOSOME
                log.debug('qc: detected chromosome replicon type via description: id=%s, description=%s', seq['id'], seq['description'])
            elif('plasmid' in sequence_description):
                seq['type'] = bc.REPLICON_PLASMID
                log.debug('qc: detected plasmid replicon type via description: id=%s, description=%s', seq['id'], seq['description'])

            sequence_desc = []
            if(cfg.keep_sequence_headers):
                if(seq['id'] in sequence_ids):
                    log.error('Fasta import: duplicated seq id! seq-id=%s', seq['id'])
                    sys.exit(f"ERROR: Detected duplicated sequence id! Sequence ID ({seq['id']}) occures multiple times!")
                else:
                    sequence_ids.add(seq['id'])
            else:
                seq['orig_id'] = seq['id']
                seq['id'] = sequence_id_generated
                seq['orig_description'] = seq['description']
                if(cfg.genus is not None or cfg.species is not None):
                    organism = ' '.join([t for t in [cfg.genus, cfg.species] if t is not None])
                    sequence_desc.append(f"[organism={organism}]")
                if(cfg.strain):
                    sequence_desc.append(f'[strain={cfg.strain}]')
                sequence_desc.append(f'[gcode={cfg.translation_table}]')

            if(seq['complete'] and seq['topology'] == bc.TOPOLOGY_CIRCULAR):  # detection of chromosomes/plasmids via sequence length thresholds
                if(seq['length'] >= bc.REPLICON_LENGTH_THRESHOLD_CHROMOSOME):
                    seq['type'] = bc.REPLICON_CHROMOSOME
                    log.debug('qc: detected replicon type via length: id=%s, type=%s, length=%i, description=%s', seq['id'], seq['type'], seq['length'], seq['description'])
                elif(seq['length'] < bc.REPLICON_LENGTH_THRESHOLD_PLASMID):
                    seq['type'] = bc.REPLICON_PLASMID
                    log.debug('qc: detected replicon type via length: id=%s, type=%s, length=%i, description=%s', seq['id'], seq['type'], seq['length'], seq['description'])
            valid_sequences.append(seq)

            if(len(sequences) == 1 and cfg.plasmid is not None):  # use plasmid mode
                seq['type'] = bc.REPLICON_PLASMID
                seq['topology'] = bc.TOPOLOGY_CIRCULAR
                seq['name'] = cfg.plasmid
            elif(replicons):  # use user provided replicon table
                sequence_id = seq['orig_id'] if 'orig_id' in seq else seq['id']
                replicon = replicons.get(sequence_id, None)
                if(replicon):
                    seq['type'] = replicon['replicon_type']
                    seq['topology'] = replicon['topology']
                    seq['complete'] = replicon['replicon_type'] != bc.REPLICON_CONTIG
                    if(replicon['name']):
                        seq['name'] = replicon['name']
                    if(not cfg.keep_sequence_headers):
                        seq['id'] = replicon['new_locus_id'] if replicon['new_locus_id'] else seq['simple_id']
                    seq.pop('simple_id')
            
            if(not cfg.keep_sequence_headers):
                if(seq['complete']):
                    sequence_desc.append('[completeness=complete]')
                if(seq['topology'] != bc.REPLICON_CONTIG):
                    sequence_desc.append(f"[topology={seq['topology']}]")
                if(seq['type'] == bc.REPLICON_CHROMOSOME):
                    sequence_desc.append('[location=chromosome]')
                elif(seq['type'] == bc.REPLICON_PLASMID):
                    if(not seq.get('name', None)):
                        seq['name'] = f'unnamed{plasmid_number}'
                        plasmid_number += 1
                    sequence_desc.append(f"[plasmid-name={seq['name']}]")
                seq['description'] = ' '.join(list(dict.fromkeys(sequence_desc)))  # remove duplicates remaining order

            if(seq['type'] == bc.REPLICON_CONTIG):
                complete_genome = False

            if(cfg.compliant):  # check INSDC compliance
                if(len(seq['id']) > 25):  # max 25 characters
                    log.error('INSDC compliance: seq id larger than 25! seq-id=%s', seq['id'])
                    sys.exit(f"ERROR: INSDC compliance failed! Sequence ID ({seq['id']}) larger than 25 characers!")
                if(bc.RE_INSDC_ID.fullmatch(seq['id']) is None):  # invalid characters
                    log.error('INSDC compliance: seq id contains invalid characters! seq-id=%s', seq['id'])
                    sys.exit(f"ERROR: INSDC compliance failed! Sequence ID ({seq['id']}) contains invalid characters!")

            log.info(
                "qc: revised sequence: id=%s, orig-id=%s, type=%s, complete=%s, topology=%s, name=%s, description='%s', orig-description='%s'",
                seq['id'], seq.get('orig_id', ''), seq['type'], seq['complete'], seq['topology'], seq.get('name', ''), seq['description'], seq.get('orig_description', '')
            )
    return valid_sequences, complete_genome


def extract_feature_sequence(feature: dict, sequence: dict) -> str:
    if(feature.get('edge', False)):
        nt = sequence['nt'][feature['start']-1:] + sequence['nt'][:feature['stop']]
    else:
        nt = sequence['nt'][feature['start']-1:feature['stop']]
    if(feature['strand'] == bc.STRAND_REVERSE):
        nt = str(Seq(nt).reverse_complement())
    return nt


def get_db_connection():
    if not hasattr(_local, "connection"):
        uri_path = cfg.db_path.joinpath('bakta.db').absolute().resolve().as_posix()
        db_uri = f"file:{uri_path}?immutable=1&mode=ro"
        # immutable=1: disables locking/coordination for NAS speed
        # mode=ro: explicitly read-only

        conn = sqlite3.connect(db_uri, uri=True, check_same_thread=False)

        # performance tuning
        conn.execute('PRAGMA journal_mode = OFF')
        conn.execute('PRAGMA cache_size = 4000')  # 4000 * 4Kb ~ 16Mb cache per thread
        conn.execute(f'PRAGMA mmap_size = {1 * 1024 * 1024 * 1024}')  # 1Gb memory mapping
        conn.row_factory = sqlite3.Row

        _local.connection = conn

    return _local.connection


import logging
import subprocess as sp
import sys

from pathlib import Path
from typing import Sequence

from Bio import SeqIO
from xopen import xopen

import bakta.config as cfg
import bakta.constants as bc
import bakta.features.orf as orf


log = logging.getLogger('EXPERT_AA_SEQ')


def search(cdss: Sequence[dict], cds_fasta_path: Path, expert_system: str, db_path: Path):
    """Conduct homology search of CDSs against PCS db."""
    diamond_output_path = cfg.tmp_path.joinpath('diamond.cds.expert.tsv')
    cmd = [
        'diamond',
        'blastp',
        '--query', str(cds_fasta_path),
        '--db', str(db_path),
        '--out', str(diamond_output_path),
        '--max-target-seqs', '1',  # single best output
        '--outfmt', '6', 'qseqid', 'sseqid', 'qlen', 'slen', 'length', 'pident', 'evalue', 'bitscore', 'stitle',
        '--threads', str(cfg.threads),
        '--load-threads', str(min(cfg.threads, 8)),  # limit load threads to max 8
        '--tmpdir', str(cfg.tmp_path),  # use tmp folder
        '--block-size', '4',  # increase block size for faster executions
        '--index-chunks', '1',  # set index chunks to 1 for faster executions
        '--sensitive'
    ]
    log.debug('cmd=%s', cmd)
    proc = sp.run(
        cmd,
        cwd=str(cfg.tmp_path),
        env=cfg.env,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        universal_newlines=True
    )
    if(proc.returncode != 0):
        log.debug('stdout=\'%s\', stderr=\'%s\'', proc.stdout, proc.stderr)
        log.warning('diamond failed! diamond-error-code=%d', proc.returncode)
        raise Exception(f'diamond error! error code: {proc.returncode}')

    cds_found = set()
    cds_by_hexdigest = orf.get_orf_dictionary(cdss)
    with diamond_output_path.open() as fh:
        for line in fh:
            (aa_identifier, model_id, query_length, model_length, alignment_length, identity, evalue, bitscore, model_title) = line.strip().split('\t')
            cds = cds_by_hexdigest[aa_identifier]
            query_cov = int(alignment_length) / len(cds['aa'])
            model_cov = int(alignment_length) / int(model_length)
            identity = float(identity) / 100
            bitscore = float(bitscore)
            evalue = float(evalue)
            (source, rank, min_identity, min_query_cov, min_model_cov, gene, product, dbxrefs) = model_title.split(' ', 1)[1].split('~~~')
            rank = int(rank)
            min_identity = float(min_identity) / 100
            min_query_cov = float(min_query_cov) / 100
            min_model_cov = float(min_model_cov) / 100
            if(query_cov >= min_query_cov and model_cov >= min_model_cov and identity >= min_identity):
                hit = {
                    'type': expert_system,
                    'source': source,
                    'rank': rank,
                    'id': model_id,
                    'gene': gene if gene != '' else None,
                    'product': product,
                    'query_cov': query_cov,
                    'subject_cov': model_cov,
                    'identity': identity,
                    'score': bitscore,
                    'evalue': evalue,
                    'db_xrefs': [] if dbxrefs == '' else dbxrefs.split(',')
                }
                if(expert_system == 'user_proteins'):
                    hit['db_xrefs'].append(f'{bc.DB_XREF_USER_PROTEINS}:{model_id}')
                cds.setdefault('expert', [])
                cds['expert'].append(hit)
                log.debug(
                    'hit: source=%s, rank=%i, seq=%s, start=%i, stop=%i, strand=%s, query-cov=%0.3f, subject-cov=%0.3f, identity=%0.3f, score=%0.1f, evalue=%1.1e, gene=%s, product=%s',
                    source, rank, cds['sequence'], cds['start'], cds['stop'], cds['strand'], query_cov, model_cov, identity, bitscore, evalue, gene, product
                )
                cds_found.add(aa_identifier)

    log.info('found=%i', len(cds_found))
    return cds_found


def write_user_protein_sequences(aa_fasta_path: Path):
    no_skipped_proteins = 0
    user_proteins = []
    try:
        with xopen(str(cfg.user_proteins), threads=0) as fh_in:
            for record in SeqIO.parse(fh_in, 'fasta'):
                try:
                    user_protein = parse_user_protein_sequences_fasta(record)
                    user_proteins.append(user_protein)
                except ValueError:
                    no_skipped_proteins += 1
    except Exception as e:
        log.error('provided user proteins file Fasta format not valid!', exc_info=True)
        sys.exit(f'ERROR: User proteins file Fasta format not valid!')

    try:
        with xopen(str(cfg.user_proteins), threads=0) as fh_in:
            for record in SeqIO.parse(fh_in, 'genbank'):
                for feature in record.features:
                    if(feature.type.lower() == 'cds'  and  bc.INSDC_FEATURE_PSEUDO not in feature.qualifiers  and  bc.INSDC_FEATURE_PSEUDOGENE not in feature.qualifiers):
                        try:
                            user_protein = parse_user_protein_sequences_genbank(feature)
                            user_proteins.append(user_protein)
                        except ValueError:
                            no_skipped_proteins += 1
                            log.warning('skipped incorrectly formatted user protein!')
    except Exception as e:
        log.error('provided user proteins file GenBank format not valid!', exc_info=True)
        sys.exit(f'ERROR: User proteins file GenBank format not valid!')

    if(len(user_proteins) > 0):
        try:
            with aa_fasta_path.open('w') as fh_out:
                for user_protein in user_proteins:
                    (model_id, min_id, min_query_cov, min_model_cov, gene, product, dbxrefs, seq) = user_protein
                    fh_out.write(f">{model_id} UserProteins~~~{101}~~~{min_id}~~~{min_query_cov}~~~{min_model_cov}~~~{gene}~~~{product}~~~{','.join(dbxrefs)}\n{seq}\n")
                    log.debug(
                        'imported user aa: id=%s, length=%i, min-id=%f, min-query-cov=%f, min-model-cov=%f, gene=%s, product=%s, dbxrefs=%s',
                        model_id, len(seq), min_id, min_query_cov, min_model_cov, gene, product, dbxrefs
                    )
        except Exception as e:
            log.error('cannot write user protein file!', exc_info=True)
            sys.exit(f'ERROR: Cannot write user protein file!')
    else:
        log.warning('no valid user proteins detected!', exc_info=True)

    return (len(user_proteins), no_skipped_proteins)


def parse_user_protein_sequences_fasta(record):
    model_id = record.id
    min_id = bc.MIN_PSC_IDENTITY * 100
    min_query_cov = bc.MIN_PSC_COVERAGE * 100
    min_model_cov = bc.MIN_PSC_COVERAGE * 100

    cols = record.description.split(' ', 1)[1].split('~~~')
    db_xrefs = []
    gene = ''
    if(len(cols) == 3):
        gene = cols[0]
        product = cols[1]
        if(cols[2] != ''):
            db_xrefs = cols[2].split(',')
    elif(len(cols) == 6):
        min_id = float(cols[0])
        min_query_cov = float(cols[1])
        min_model_cov = float(cols[2])
        gene = cols[3]
        product = cols[4]
        if(cols[5] != ''):
            db_xrefs = cols[5].split(',')
    else:
        log.warning('wrong description format in FASTA user protein file! id=%s, description=%s', record.id, record.description)
        raise ValueError(f'Wrong description format in FASTA user protein file!')

    gene = gene.strip()
    product = product.strip()
    if(product == ''):
        log.warning('missing product in FASTA user protein file! id=%s, description=%s', record.id, record.description)
        raise ValueError('Missing product in FASTA user protein file!')

    for db_xref in db_xrefs:
        if(':' not in db_xref):
            log.warning('wrong dbxref format in FASTA user protein file! id=%s, dbxref=%s', record.id, db_xref)
            raise ValueError(f'Wrong dbxref format in FASTA user protein file!')
    seq = str(record.seq).upper()
    return (model_id, min_id, min_query_cov, min_model_cov, gene, product, db_xrefs, seq)


def parse_user_protein_sequences_genbank(feature: dict):
    min_id = bc.MIN_PSC_IDENTITY * 100
    min_query_cov = bc.MIN_PSC_COVERAGE * 100
    min_model_cov = bc.MIN_PSC_COVERAGE * 100

    q = feature.qualifiers
    model_ids = q.get('locus_tag', [])
    model_id = model_ids[0] if len(model_ids) > 0 else None
    if(model_id is None):
        log.warning('missing locus_tag in GenBank user protein file!')
        raise ValueError('Missing locus_tag in GenBank user protein file!')

    genes = q.get('gene', [])
    gene = genes[0].strip() if len(genes) > 0 else ''
    
    products = q.get('product', [])
    product = products[0].strip() if len(products) > 0 else None
    if(product is None or product == ''):
        log.warning('missing product in GenBank user protein file!')
        raise ValueError('Missing product in GenBank user protein file!')

    db_xrefs = q.get('db_xref', [])
    for db_xref in db_xrefs:
        if(':' not in db_xref):
            log.warning('wrong dbxref format in Fasta user protein file! dbxref=%s', db_xref)
            raise ValueError(f'Wrong dbxref format in Fasta user protein file! dbxref={db_xref}')
    ecs = q.get('EC_number', [])
    for ec in ecs:
        db_xrefs.append(f'EC:{ec}')

    seqs = q.get('translation', [])
    seq = str(seqs[0]).upper() if len(seqs) > 0 else None
    if(seq is None):
        log.warning('missing sequence translation in GenBank user protein file!')
        raise ValueError('Missing sequence translation in GenBank user protein file!')

    return (model_id, min_id, min_query_cov, min_model_cov, gene, product, db_xrefs, seq)
